#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Projet for spatial audio separation
from this paper:

@inproceedings{fitzgeraldPROJETa,
TITLE = {{PROJET - Spatial Audio Separation Using Projections}},
AUTHOR = {D. Fitzgerald and A. Liutkus and R. Badeau},
BOOKTITLE = {{41st International Conference on Acoustics, Speech and Signal Processing (ICASSP)}},
ADDRESS = {Shanghai, China},
PUBLISHER = {{IEEE}},
YEAR = {2016},
}

Copyright (c) 2016, Antoine Liutkus, Inria

modified by Ethan Manilow and Prem Seetharaman for incorporation into nussl.
"""

import numpy as np

import separation_base
from ..core import utils
from ..core import constants
from ..core.audio_signal import AudioSignal


class Projet(separation_base.SeparationBase):
    """Implements foreground/background separation using the 2D Fourier Transform

    Parameters:
        input_audio_signal: (AudioSignal object) The AudioSignal object that has the
                            audio data that REPET will be run on.
        use_librosa_stft: (Optional) (bool) Calls librosa's stft function instead of nussl's

    """
    def __init__(self, input_audio_signal, num_sources,
                 num_iterations=200, num_panning_directions=41, num_projections=15,
                 matrix_datatype='float32', panning_profiles=30,
                 verbose=False, use_librosa_stft=constants.USE_LIBROSA_STFT):
        super(Projet, self).__init__(input_audio_signal=input_audio_signal)
        
        if not self.audio_signal.is_stereo:
            raise ValueError('Can only run PROJET on a stereo audio signal!')

        self.num_sources = num_sources
        self.num_iterations = num_iterations
        self.num_panning_directions = num_panning_directions
        self.num_projections = num_projections
        self.panning_profiles = panning_profiles

        if isinstance(matrix_datatype, str):
            matrix_datatype = np.dtype(matrix_datatype)

        if not np.issubdtype(matrix_datatype, np.float):
            raise ValueError('matrix_datatype must be a float!')

        self.matrix_datatype = matrix_datatype

        self.verbose = verbose

        self.stft = None
        self.sources = None
        self.use_librosa_stft = use_librosa_stft

    def run(self):
        """

        Returns:
            sources (list of AudioSignals): A list of AudioSignal objects with all of the sources found in the mixture

        Example:
             ::

        """
        self._compute_spectrograms()
        
        (num_freq_bins, num_time_bins, num_channels) = self.stft.shape
        num_sources = self.num_sources
        eps = 1e-20

        if self.verbose: print('Initializing panning matrix...')
        # initialize PSD and panning to random
        P = np.abs(np.random.randn(num_freq_bins * num_time_bins, num_sources)).astype(self.matrix_datatype) + 1

        # panning_sources_matrix is number of panning directions
        # to look for by number of sources (Q in original paper)
        panning_sources_matrix = np.abs(np.random.randn(self.num_panning_directions,
                                                        num_sources)).astype(self.matrix_datatype) + 1

        chan_pan_diff = utils.complex_randn((num_channels, self.num_panning_directions - self.panning_profiles))
        chan_per_panning_profiles = self.multichannel_grid(num_channels, self.panning_profiles)

        if self.verbose: print('Computing initial panning profiles...')

        # compute panning profiles
        panning_matrix = np.concatenate((chan_pan_diff, chan_per_panning_profiles), axis=1)
        panning_matrix /= np.sqrt(np.sum(np.abs(panning_matrix) ** 2, axis=0))[None, ...]

        if self.verbose: print('Computing initial projection matrices...')

        # compute projection matrix
        projection_matrix = np.concatenate((utils.complex_randn((max(self.num_projections - 5, 0), num_channels)),
                                            self.orthogonal_matrix(self.multichannel_grid(num_channels, min(self.num_projections, 5)))))
        projection_matrix /= np.sqrt(np.sum(np.abs(projection_matrix) ** 2, axis=1))[..., None]

        if self.verbose: print('Computing K matrix.')
        # compute K matrix
        K = np.abs(np.dot(projection_matrix, panning_matrix)).astype(np.float32)

        if self.verbose: print('Computing projections and storing spectrograms and squared spectrograms.')
        # compute the projections and store their spectrograms and squared spectrograms
        C = np.tensordot(self.stft, projection_matrix, axes=(2, 1))
        C = np.reshape(C, (num_freq_bins * num_time_bins, self.num_projections))

        # NOTE: C now the same shape as P.
        V = np.abs(C).astype(np.float32)
        V2 = V ** 2

        # noinspection PyUnusedLocal
        C = []  # release memory

        if self.verbose: print('Starting iterations')
        # main iterations
        for iteration in range(self.num_iterations):

            if self.verbose:
                print('Iteration {}'.format(iteration))

            sigma = np.dot(P, np.dot(panning_sources_matrix.T, K.T))

            if self.verbose: print('\tUpdating P...')
            # updating P
            P *= np.dot(1.0 / (sigma + eps), np.dot(K, panning_sources_matrix)) / \
                 (np.dot(3 * sigma / (sigma ** 2 + V2 + eps), np.dot(K, panning_sources_matrix)))

            if self.verbose: print('\tUpdating sigma')
            # the following line is an optional trick that enforces orthogonality of the spectrograms.
            # P*=(100+P)/(100+np.sum(P,axis=1)[...,None])
            # update sigma using updated P. transpose to fit into Q. (15, F*T)
            sigma = np.dot(P, np.dot(panning_sources_matrix.T, K.T)).T

            if self.verbose: print('\tUpdating panning sources matrix')
            # updating Q
            panning_sources_matrix *= np.dot(K.T, np.dot(np.divide(1.0, sigma + eps), P)) / \
                 np.dot(K.T, np.dot(np.divide(3 * sigma,  (sigma ** 2 + V2.T + eps)), P))

        if self.verbose: print('Completing final separation')
        # final separation
        recompose_matrix = np.linalg.pinv(projection_matrix)  # IxM

        sigma = np.dot(P, np.dot(panning_sources_matrix.T, K.T))
        C = np.dot(np.reshape(self.stft, (num_freq_bins * num_time_bins, num_channels)), projection_matrix.T)

        self.sources = []

        if self.verbose: print('Making AudioSignal objects')
        for j in range(num_sources):
            sigma_j = np.outer(P[:, j], np.dot(panning_sources_matrix[:, j].T, K.T))
            source_stft = sigma_j / sigma * C
            source_stft = np.dot(source_stft, recompose_matrix.T)
            source_stft = np.reshape(source_stft, (num_freq_bins, num_time_bins, num_channels))
            source = AudioSignal(stft=source_stft, sample_rate=self.audio_signal.sample_rate)
            source.istft(self.stft_params.window_length, self.stft_params.hop_length, 
                        self.stft_params.window_type, overwrite=True, 
                        use_librosa=self.use_librosa_stft, 
                        truncate_to_length=self.audio_signal.signal_length)
            self.sources.append(source)

        if self.verbose: print('Projet finished running.')
        return self.sources
    
    def _compute_spectrograms(self):
        self.stft = self.audio_signal.stft(overwrite=True, remove_reflection=True, use_librosa=self.use_librosa_stft)

    def multichannel_grid(self, I, L, sigma=1, normalize=True):
        # 15 points equally spaced between 0 and num_channels - 1 (1). 15 points between 0 and 1 basically.
        pos = np.linspace(0, I - 1, L)
        # 2 by 15 all 0s.
        res = np.zeros((I, L))
        for i in range(I):
            # each row becomes e^(+/-[0, 1]**2)
            res[i, ...] = np.exp(-(pos - i) ** 2 / sigma ** 2)
        if normalize:
            res /= np.sqrt(np.sum(res ** 2, axis=0))
        return res

    def orthogonal_matrix(self, R):
        # 2 by 15
        (I, L) = R.shape
        # 15 by 2
        res = np.ones((L, I))
        # all rows, squeeze removes all one dimensional entries (1, 3, 1) shape goes to (3) shape.
        # sum of all rows of R but the last one, along each column.
        # all columns of res but the last one become the last row of R (2 by 1)
        # divided by the sum of all the columns of R but the last one.
        # transpose to fit into res.
        res[:, -1] = - (R[-1, :] / np.squeeze(np.sum(R[:-1, :], axis=0))).T
        # normalize res by rms along each row
        res /= np.sqrt(np.sum(res ** 2, axis=1))[..., None]
        return res
    
    def make_audio_signals(self):
        """ Returns the background and foreground audio signals. You must have run FT2D.run() prior
        to calling this function. This function will return None if run() has not been called.

        Returns:
            Audio Signals (List):  element list.

        EXAMPLE:
             ::
        """

        return self.sources

    def plot(self, output_name, **kwargs):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
A repository containing all of the constants frequently used in
this wacky, mixed up source separation stuff.
"""
import scipy.signal
from six.moves.urllib_parse import urljoin

__all__ = ['DEFAULT_SAMPLE_RATE', 'DEFAULT_WIN_LEN_PARAM', 'DEFAULT_BIT_DEPTH',
           'DEFAULT_MAX_VAL', 'EPSILON', 'MAX_FREQUENCY',
           'WINDOW_HAMMING', 'WINDOW_RECTANGULAR', 'WINDOW_HANN',
           'WINDOW_BLACKMAN', 'WINDOW_TRIANGULAR', 'WINDOW_DEFAULT',
           'ALL_WINDOWS', 'NUMPY_JSON_KEY', 'LEN_INDEX', 'CHAN_INDEX',
           'STFT_VERT_INDEX', 'STFT_LEN_INDEX', 'STFT_CHAN_INDEX']

DEFAULT_SAMPLE_RATE = 44100  #: (int): Default sample rate. 44.1 kHz, CD-quality
DEFAULT_WIN_LEN_PARAM = 0.04  #: (float): Default window length. 40ms
DEFAULT_WIN_LENGTH = 2048  #: (int): Default window length, 2048 samples.
DEFAULT_BIT_DEPTH = 16  #: (int): Default bit depth. 16-bits, CD-quality
DEFAULT_MAX_VAL = 2 ** 16  #: (int): Max value of 16-bit audio file (unsigned)
EPSILON = 1e-16  #: (float): epsilon for determining small values
MAX_FREQUENCY = DEFAULT_SAMPLE_RATE // 2  #: (int): Maximum frequency representable. 22050 Hz

WINDOW_HAMMING = scipy.signal.hamming.__name__  #: (str): Name for calling Hamming window. 'hamming'
WINDOW_RECTANGULAR = 'rectangular'  #: (str): Name for calling Rectangular window. 'rectangular'
WINDOW_HANN = scipy.signal.hann.__name__  #: (str): Name for calling Hann window. 'hann'
WINDOW_BLACKMAN = scipy.signal.blackman.__name__  #: (str): Name for calling Blackman window. 'blackman'
WINDOW_TRIANGULAR = 'triangular'  #: (str): Name for calling Triangular window. 'triangular'

WINDOW_DEFAULT = WINDOW_HAMMING  #: (str): Default window, Hamming.
ALL_WINDOWS = [WINDOW_HAMMING, WINDOW_RECTANGULAR, WINDOW_HANN, WINDOW_BLACKMAN, WINDOW_TRIANGULAR]
"""list(str): list of all available windows in *nussl*
"""

NUMPY_JSON_KEY = "py/numpy.ndarray"  #: (str): key used when turning numpy arrays into json

BINARY_MASK = 'binary'
""" String alias for setting this object to return :class:`separation.masks.binary_mask.BinaryMask` objects
"""

SOFT_MASK = 'soft'
""" String alias for setting this object to return :class:`separation.masks.soft_mask.SoftMask` objects
"""

# ############# Array Indices ############# #

# audio_data
LEN_INDEX  = 1  #: (int): Index of the number of samples in an audio signal. Used in :ref:`audio_signal`
CHAN_INDEX = 0  #: (int): Index of the number of channels in an audio signal. Used in :ref:`audio_signal`

# stft_data
STFT_VERT_INDEX = 0
"""
(int) Index of the number of frequency (vertical) values in a time-frequency representation. 
Used in :ref:`audio_signal` and in :ref:`mask_base`.
"""
STFT_LEN_INDEX  = 1
"""
(int) Index of the number of time (horizontal) hops in a time-frequency representation. 
Used in :ref:`audio_signal` and in :ref:`mask_base`.
"""
STFT_CHAN_INDEX = 2
"""
(int) Index of the number of channels in a time-frequency representation. 
Used in :ref:`audio_signal` and in :ref:`mask_base`.
"""

# ############# nussl-extras urls ############# #

NUSSL_EFZ_BASE_URL = 'http://nussl.ci.northwestern.edu/'
NUSSL_EFZ_STATIC_BASE_URL = urljoin(NUSSL_EFZ_BASE_URL, 'static/')
NUSSL_EFZ_AUDIO_URL = urljoin(NUSSL_EFZ_STATIC_BASE_URL, 'audio/')
NUSSL_EFZ_MODELS_URL = urljoin(NUSSL_EFZ_STATIC_BASE_URL, 'trained_models/')
NUSSL_EFZ_BENCHMARKS_URL = urljoin(NUSSL_EFZ_STATIC_BASE_URL, 'benchmarks/')
NUSSL_EFZ_AUDIO_METADATA_URL = urljoin(NUSSL_EFZ_BASE_URL, 'audio-metadata.json')
NUSSL_EFZ_BENCHMARK_METADATA_URL = urljoin(NUSSL_EFZ_BASE_URL, 'benchmark-metadata.json')
NUSSL_EFZ_MODEL_METADATA_URL = urljoin(NUSSL_EFZ_BASE_URL, 'model-metadata.json')

USE_LIBROSA_STFT = False  #: (bool): Whether *nussl* will use librosa's stft function by default


# ############# MUSDB interface ############### #
STEM_TARGET_DICT = {'vocals': {'vocals': 1},
                    'drums': {'drums': 1},
                    'bass': {'bass': 1},
                    'other': {'other': 1},
                    'accompaniment': {'bass': 1, 'drums': 1, 'other': 1}}

VOX_ACC_DICT = {'vocals': {'vocals': 1}, 'accompaniment': {'accompaniment': 1}}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Implements a simple high/low pass filter for audio source separation
"""

from __future__ import division
import numpy as np
import scipy.signal

import mask_separation_base
from ideal_mask import IdealMask


class HighLowPassFilter(mask_separation_base.MaskSeparationBase):
    """
    A simple high/low pass filter that creates a mask in the time frequency representation
    """

    def __init__(self, input_audio_signal, high_pass_cutoff_hz, do_fir_filter=False,
                 force_recompute_stft=False,
                 mask_type=mask_separation_base.MaskSeparationBase.BINARY_MASK):
        super(HighLowPassFilter, self).__init__(input_audio_signal=input_audio_signal,
                                                mask_type=mask_type)
        self.high_pass_cutoff_hz = high_pass_cutoff_hz

        self.should_do_fir_filter = do_fir_filter
        self.do_stft = force_recompute_stft
        self.stft = None

        self.high_pass_mask = None
        self.low_pass_mask = None

        self.high_pass_signal = None
        self.low_pass_signal = None

    def run(self):
        """

        Returns:

        """

        if self.should_do_fir_filter:

            # This implementation is based off of the sci-py cookbook:
            # https://scipy-cookbook.readthedocs.io/items/FIRFilter.html

            # Set up filter parameters
            nyquist_rate = self.audio_signal.sample_rate / 2
            width = 5.0 / nyquist_rate  # Hz
            ripple_db = 60

            # Convert cutoff frequency
            cutoff_hz = self.high_pass_cutoff_hz / nyquist_rate  # Hz

            # Make the filter
            filter_order, beta = scipy.signal.kaiserord(ripple_db, width)
            filter_coeffs = scipy.signal.firwin(filter_order, cutoff_hz, window=('kaiser', beta))

            # Apply the filter to every channel in the mixture
            low_pass_array = []
            for ch in self.audio_signal.get_channels():
                low_pass_array.append(scipy.signal.lfilter(filter_coeffs, [1.0], ch))

            # Make a new AudioSignal object with filtered signal
            low_pass_array = np.array(low_pass_array)
            self.low_pass_signal = self.audio_signal.make_copy_with_audio_data(low_pass_array,
                                                                               verbose=False)
            self.high_pass_signal = self.audio_signal - self.low_pass_signal

            # Make masks
            ideal_mask = IdealMask(self.audio_signal, [self.high_pass_signal, self.low_pass_signal],
                                   mask_type=self.mask_type)
            self.high_pass_mask, self.low_pass_mask = ideal_mask.run()

        else:
            # This is the more simple of the two filtering methods. Here we just zero out STFT bins

            # Compute the spectrogram and find the closest frequency bin to the cutoff freq
            self._get_stft()
            closest_freq_bin = self.audio_signal.get_closest_frequency_bin(self.high_pass_cutoff_hz)

            # Make masks
            self.low_pass_mask = self.ones_mask(self.stft.shape)
            self.low_pass_mask.mask[closest_freq_bin:, :, :] = 0

            self.high_pass_mask = self.low_pass_mask.invert_mask()

        self.result_masks = [self.low_pass_mask, self.high_pass_mask]

        return self.result_masks

    def _get_stft(self):
        """
        Computes the spectrogram of the input audio signal
        Returns:

        """
        if not self.audio_signal.has_stft_data or self.do_stft:
            self.stft = self.audio_signal.stft(overwrite=True)
        else:
            self.stft = self.audio_signal.stft_data

    def make_audio_signals(self):
        """

        Returns:

        """
        if self.low_pass_mask is None or self.high_pass_mask is None:
            raise ValueError('Must call run() before calling make_audio_signals()!')

        if not self.should_do_fir_filter:
            self.high_pass_signal = self.audio_signal.apply_mask(self.high_pass_mask)
            self.high_pass_signal.istft(overwrite=True,
                                        truncate_to_length=self.audio_signal.signal_length)

            self.low_pass_signal = self.audio_signal.apply_mask(self.low_pass_mask)
            self.low_pass_signal.istft(overwrite=True,
                                       truncate_to_length=self.audio_signal.signal_length)

        return self.low_pass_signal, self.high_pass_signal



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Wrapper class for ``museval`` implementation of the BSS-Eval metrics (SDR, SIR, SAR).
Contains logic for loading ground truth AudioSignals and estimated
AudioSignals to compute BSS-Eval metrics. The ``mir_eval`` module contains an
implementation of BSS-Eval version 4.
"""
import museval
import json

import bss_eval_base
from ..core import constants, utils


class BSSEvalV4(bss_eval_base.BSSEvalBase):
    """
    Wrapper class for ``museval`` implementation of the BSS-Eval metrics (SDR, SIR, SAR).
    Contains logic for loading ground truth AudioSignals and estimated
    AudioSignals to compute BSS-Eval metrics. The ``mir_eval`` module contains an
    implementation of BSS-Eval version 4.

    The BSS-Eval metrics attempt to measure perceptual quality by comparing sources
    estimated from a source separation algorithm to the ground truth, known sources.
    These metrics evaluate the distortion (SDR) and artifacts (SAR) present in the
    estimated signals as well as the interference (SIR) from other sources in a given
    estimated source. Results are returned in units of dB, with higher values indicating
    better quality.

    Examples:
        :class:`BSSEvalV4` can be initialized in two ways, either with a list or a dict.
        See the example below for a demonstration:

        .. code-block:: python
            :linenos:

            mir1k_dir = 'path/to/MIR-1K'
            mix, vox, acc = next(nussl.datasets.mir1k(mir1k_dir))
            mix.to_mono(overwrite=True)

            r = nussl.RepetSim(mix)
            r()
            bg, fg = r.make_audio_signals()

            # Method 1: Dictionary where sources are explicit
            # Note that this dictionary exactly matches nussl.constants.VOX_ACC_DICT
            est_dict = {'vocals': fg, 'accompaniment': bg}
            gt_dict = {'vocals': vox, 'accompaniment': acc}
            bss = nussl.evaluation.BSSEvalV4(mix, gt_dict, est_dict)
            scores1 = bss.evaluate()

            # Method 2: List
            # Note that vocals are always expected to be first, then accompaniment.
            bss = nussl.evaluation.BSSEvalV4(mix, [vox, acc], [fg, bg])
            scores2 = bss.evaluate()

    See Also:
        * For more information on ``museval`` (python implementation of BSS-Eval v4) see
        `its Github page <https://github.com/sigsep/sigsep-mus-eval>`_ or
        'its documentation <https://sigsep.github.io/sigsep-mus-eval/>`_.

        * For more information on the BSS-Eval metrics, see the webpage for
        `the original MATLAB implementation <http://bass-db.gforge.inria.fr/bss_eval/>`_.

        * :class:`BSSEvalSources` and :class:`BSSEvalImages` for the ``mir_eval`` version 3
        BSS-Eval implementations.

    References:
        * Emmanuel Vincent, Rémi Gribonval, Cédric Févotte. Performance measurement in blind
        audio source separation. IEEE Transactions on Audio, Speech and Language Processing,
        Institute of Electrical and Electronics Engineers, 2006, 14 (4), pp.1462–1469.
        <inria-00544230>

        * Fabian-Robert Stöter, Antoine Liutkus, and Nobutaka Ito. The 2018 Signal Separation
        Evaluation Campaign. In International Conference on Latent Variable Analysis and Signal
        Separation, pages 293–305. Springer, 2018.

    Args:
        true_sources_list (list): List of :class:`AudioSignal` objects that contain the ground
            truth sources for the mixture.
        estimated_sources_list (list):  List of :class:`AudioSignal` objects that contain estimate
            sources, output from source separation algorithms.
        source_labels (list): List of strings that are labels for each source to be used as keys for
            the scores. Default value is ``None`` and in that case labels are ``Source 0``,
            ``Source 1``, etc.
        algorithm_name (str): Name of the algorithm if using this object to compute many
            BSS-Eval metrics. Can be changed later.
        do_mono (bool): Should flatten the audio to mono before calculating metrics.
        compute_permutation (bool): Should try to find the best permutation for the estimated
            sources.
    """
    SDR_MEANS = 'sdr_means'

    def __init__(self, mixture, true_sources, estimated_sources,
                 target_dict=constants.VOX_ACC_DICT,
                 mode='v4', output_dir=None, win=1.0, hop=1.0):
        # try:
        #     super(BSSEvalV4, self).__init__(true_sources_list=true_sources_list,
        #                                     estimated_sources_list=estimated_sources_list,
        #                                     source_labels=source_labels, do_mono=do_mono)
        # except evaluation_base.AudioSignalListMismatchError:
        #     pass

        # if vox_acc and target_dict == constants.VOX_ACC_DICT:
        #     self.source_labels = ['accompaniment', 'vocals']
        #
        # if target_dict == constants.STEM_TARGET_DICT:
        #     self.source_labels = ['drums', 'bass', 'other', 'vocals']
        #
        # self.true_sources = {l: self.true_sources_list[i] for i, l in enumerate(self.source_labels)}
        # self.estimates = {l: self.estimated_sources_list[i].audio_data.T
        #                        for i, l in enumerate(self.source_labels)}
        if type(true_sources) is not type(estimated_sources):
            raise bss_eval_base.BssEvalException('true_sources and estimated_sources must both be '
                                                 'lists or both be dicts!')

        have_list = type(true_sources) is list

        self._scores = {}
        self.target_dict = target_dict
        self.is_vox_acc = target_dict == constants.VOX_ACC_DICT
        self.is_stem = target_dict == constants.STEM_TARGET_DICT
        self.mixture = mixture
        self.mode = mode
        self.output_dir = output_dir
        self.win = win
        self.hop = hop

        # Set up the dictionaries for museval
        # self.true_sources is filled with AudioSignals (b/c nussl converts it to a Track)
        # & self.estimates is raw numpy arrays
        if self.is_vox_acc:
            if have_list:
                self.estimates = {'vocals': estimated_sources[0].audio_data.T,
                                  'accompaniment': estimated_sources[1].audio_data.T}
                self.true_sources = {'vocals': true_sources[0],
                                     'accompaniment': true_sources[1]}
            else:
                self.estimates = {'vocals': estimated_sources['vocals'].audio_data.T,
                                  'accompaniment': estimated_sources['accompaniment'].audio_data.T}
                self.true_sources = {'vocals': true_sources['vocals'],
                                     'accompaniment': true_sources['accompaniment']}
        else:
            # Assume they know what they're doing...
            self.true_sources = true_sources
            self.estimates = estimated_sources

            # TODO: STEM_TARGET_DICT logic

    def _get_scores(self, scores):
        s = scores.split()
        v, a = s[0], s[6].replace('\\n', '')
        i1, i2 = [2, 3, 4, 5], [8, 9, 10, 11]
        return {v: {self._parse(s[i])[0]: self._parse(s[i])[1] for i in i1},
                a: {self._parse(s[i])[0]: self._parse(s[i])[1] for i in i2}}

    @staticmethod
    def _parse(str_):
        bss_type, val = str_.split(':')
        val = float(val[:-3])
        return bss_type, val

    def _get_mean_scores(self, scores):
        return self._get_scores(repr(scores))

    def evaluate(self):
        track = utils.audio_signals_to_musdb_track(self.mixture, self.true_sources,
                                                   self.target_dict)

        bss_output = museval.eval_mus_track(track, self.estimates,
                                            output_dir=self.output_dir, mode=self.mode,
                                            win=self.win, hop=self.hop)

        self._populate_scores_dict(bss_output)

        return self.scores

    def _populate_scores_dict(self, bss_output):

        self.scores[self.RAW_VALUES] = json.loads(bss_output.json)  # Hack to format dict correctly
        self.scores[self.SDR_MEANS] = self._get_mean_scores(repr(bss_output))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import warnings

import numpy as np
from scipy.ndimage.filters import convolve
import vamp

from ..core import constants
import mask_separation_base
import masks
from .. import AudioSignal


class Melodia(mask_separation_base.MaskSeparationBase):
    """Implements melody extraction using Melodia.

    J. Salamon and E. Gómez, "Melody Extraction from Polyphonic Music Signals using Pitch Contour Characteristics",
    IEEE Transactions on Audio, Speech and Language Processing, 20(6):1759-1770, Aug. 2012.

    This needs melodia installed as a vamp plugin, as well as having vampy= for Python installed.

    Parameters:
        input_audio_signal: (AudioSignal object) The AudioSignal object that has the
                            audio data that REPET will be run on.
        high_pass_cutoff: (Optional) (float) value (in Hz) for the high pass cutoff filter.
        do_mono: (Optional) (bool) Flattens AudioSignal to mono before running the algorithm (does not effect the
                        input AudioSignal object)
        use_librosa_stft: (Optional) (bool) Calls librosa's stft function instead of nussl's

    """

    def __init__(self, input_audio_signal, high_pass_cutoff=None, minimum_frequency=55.0,
                 maximum_frequency=1760.0, voicing_tolerance=0.5, minimum_peak_salience=0.0,
                 do_mono=False, use_librosa_stft=constants.USE_LIBROSA_STFT,
                 mask_type=constants.SOFT_MASK, mask_threshold=0.5):

        super(Melodia, self).__init__(input_audio_signal=input_audio_signal, 
                                      mask_type=mask_type, mask_threshold=mask_threshold)
        self.high_pass_cutoff = 100.0 if high_pass_cutoff is None else float(high_pass_cutoff)
        self.background = None
        self.foreground = None
        self.use_librosa_stft = use_librosa_stft
        self.minimum_frequency = float(minimum_frequency)
        self.maximum_frequency = float(maximum_frequency)
        self.voicing_tolerance = float(voicing_tolerance)
        self.minimum_peak_salience = float(minimum_peak_salience)
        self.stft = None
        self.melody = None
        self.melody_signal = None
        self.timestamps = None
        self.foreground_mask = None
        self.background_mask = None

        if do_mono:
            self.audio_signal.to_mono(overwrite=True)

    def extract_melody(self):
        params = {}
        params['minfqr'] = self.minimum_frequency
        params['maxfqr'] = self.maximum_frequency
        params['voicing'] = self.voicing_tolerance
        params['minpeaksalience'] = self.minimum_peak_salience

        try:
            data = vamp.collect(self.audio_signal.audio_data, self.sample_rate,
                                "mtg-melodia:melodia", parameters=params)
        except Exception as e:
            print('**~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~**\n'
                  '*          Are Vamp and Melodia installed correctly?          *\n'
                  '* Check https://bit.ly/2DXbrAk for installation instructions! *\n'
                  '**~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~**')
            raise e

        _, melody = data['vector']
        hop = 128./44100. # hard coded hop in Melodia vamp plugin, converting it to frames.
        timestamps = 8 * hop + np.arange(len(melody)) * hop
        melody[melody < 0] = 0
        self.melody = melody
        self.timestamps = timestamps

    def create_melody_signal(self, num_overtones):
        # Adapted from Melosynth by Justin Salamon: https://github.com/justinsalamon/melosynth

        if self.timestamps[0] > 0:
            estimated_hop = np.median(np.diff(self.timestamps))
            previous_time = max(self.timestamps[0] - estimated_hop, 0)
            self.timestamps = np.insert(self.timestamps, 0, previous_time)
            self.melody = np.insert(self.melody, 0, 0)

        sample_rate = self.audio_signal.sample_rate
        melody_signal = []
        transition_length = .010 # duration for fade in/out and frequency interpretation
        phase = np.zeros(num_overtones)
        previous_frequency = 0
        previous_time = 0

        for time, frequency in zip(self.timestamps, self.melody):
            num_samples = int(np.round((time - previous_time) * sample_rate))
            if num_samples > 0:
                num_transition_samples = float(min(np.round(transition_length * sample_rate),
                                                   num_samples))
                frequency_series = np.ones(num_samples) * previous_frequency

                if previous_frequency > 0 and frequency > 0:
                    frequency_series += np.minimum(np.arange(num_samples) / num_transition_samples, 1) * \
                                        (frequency - previous_frequency)
                elif frequency > 0:
                    frequency_series = np.ones(num_samples) * frequency

                samples = np.zeros(num_samples)
                for overtone in range(num_overtones):
                    overtone_num = overtone + 1
                    phasors = 2 * np.pi * overtone_num * frequency_series / float(sample_rate)
                    phases = phase[overtone] + np.cumsum(phasors)
                    samples += np.sin(phases) / overtone_num
                    phase[overtone] = phases[-1]

                if previous_frequency == 0 and frequency > 0:
                    samples *= np.minimum(np.arange(num_samples) / num_transition_samples, 1)
                elif previous_frequency > 0 and frequency == 0:
                    samples *= np.maximum(1 - np.arange(num_samples) / num_transition_samples, 0)
                elif previous_frequency == 0 and frequency == 0:
                    samples *= 0

                melody_signal.extend(samples)

            previous_frequency = frequency
            previous_time = time

        melody_signal = np.asarray(melody_signal)
        melody_signal *= 0.8 / float(np.max(melody_signal))
        melody_signal = [melody_signal for channel in range(self.audio_signal.num_channels)]
        melody_signal = np.asarray(melody_signal)
        melody_signal = melody_signal[:, 0:self.audio_signal.signal_length]
        melody_signal = AudioSignal(audio_data_array=melody_signal, sample_rate=sample_rate)

        self.melody_signal = melody_signal
        return melody_signal

    def create_harmonic_mask(self, melody_signal):
        normalized_melody_stft = np.abs(melody_signal.stft())
        normalized_melody_stft /= np.max(normalized_melody_stft)

        # Need to threshold the melody stft since the synthesized
        # F0 sequence overtones are at different weights.
        normalized_melody_stft = normalized_melody_stft > 1e-2
        normalized_melody_stft = normalized_melody_stft.astype(float)
        mask = np.empty(self.audio_signal.stft().shape)

        # Smoothing the mask row-wise using a low-pass filter to
        # get rid of discontuinities in the mask.
        kernel =  np.full((1, 20), 1/20.)
        for channel in range(self.audio_signal.num_channels):
            mask[:, :, channel] = convolve(normalized_melody_stft[:, :, channel], kernel)
        return mask

    def run(self):
        """

        Returns:
            foreground (AudioSignal): An AudioSignal object with melodic foreground in
            foreground.audio_data
            (to get the corresponding background run self.make_audio_signals())

        Example:
             ::

        """
        # High pass filter cutoff freq. (in # of freq. bins), +1 to match MATLAB implementation
        self.high_pass_cutoff = int(np.ceil(self.high_pass_cutoff * (self.stft_params.n_fft_bins - 1) /
                                            self.audio_signal.sample_rate)) + 1
        self._compute_spectrum()

        # separate the mixture foreground melody by masking
        if self.melody_signal is None:
            self.extract_melody()
            self.create_melody_signal(100)

        foreground_mask = self.create_harmonic_mask(self.melody_signal)
        foreground_mask[0:self.high_pass_cutoff, :] = 0

        foreground_mask = masks.SoftMask(foreground_mask)
        if self.mask_type == self.BINARY_MASK:
            foreground_mask = foreground_mask.mask_to_binary(self.mask_threshold)

        self.foreground_mask = foreground_mask
        self.background_mask = foreground_mask.invert_mask()

        self.foreground = self.audio_signal.apply_mask(foreground_mask)
        self.foreground.istft(self.stft_params.window_length, self.stft_params.hop_length,
                              self.stft_params.window_type,
                              overwrite=True, use_librosa=self.use_librosa_stft,
                              truncate_to_length=self.audio_signal.signal_length)

        return [self.background_mask, self.foreground_mask]

    def _compute_spectrum(self):
        self.stft = self.audio_signal.stft(overwrite=True, remove_reflection=True,
                                           use_librosa=self.use_librosa_stft)

    def make_audio_signals(self):
        """ Returns the background and foreground audio signals. You must have run FT2D.run() prior
        to calling this function. This function will return None if run() has not been called.

        Returns:
            Audio Signals (List): 2 element list.

                * bkgd: Audio signal with the calculated background track
                * fkgd: Audio signal with the calculated foreground track

        EXAMPLE:
             ::
        """
        if self.foreground is None:
            return None

        background_array = self.audio_signal.audio_data - self.foreground.audio_data
        self.background = self.audio_signal.make_copy_with_audio_data(background_array)
        return [self.background, self.foreground]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
The REpeating Pattern Extraction Technique using the Similarity Matrix (REPET-SIM)
"""

import warnings

import numpy as np

import mask_separation_base
import masks
from ..core import utils
from ..core import constants


class RepetSim(mask_separation_base.MaskSeparationBase):
    """Implements the REpeating Pattern Extraction Technique algorithm using the Similarity Matrix (REPET-SIM).

    REPET is a simple method for separating the repeating background from the non-repeating foreground in a piece of
    audio mixture. REPET-SIM is a generalization of REPET, which looks for similarities instead of periodicities.

    References:

        * Zafar Rafii and Bryan Pardo. "Audio Separation System and Method," US20130064379 A1, US 13/612,413, March 14,
          2013
        * Zafar Rafii and Bryan Pardo. "Music/Voice Separation using the Similarity Matrix," 13th International Society
          on Music Information Retrieval, Porto, Portugal, October 8-12, 2012.

    See Also:
        http://music.eecs.northwestern.edu/research.php?project=repet

    Parameters:

    Examples:
        :ref:`The RepetSim Demo Example <repet_sim_demo>`
    """

    def __init__(self, input_audio_signal, similarity_threshold=None, min_distance_between_frames=None,
                 max_repeating_frames=None, high_pass_cutoff=None, do_mono=False,
                 use_librosa_stft=constants.USE_LIBROSA_STFT, matlab_fidelity=False,
                 mask_type=mask_separation_base.MaskSeparationBase.SOFT_MASK, mask_threshold=0.5):
        super(RepetSim, self).__init__(input_audio_signal=input_audio_signal, mask_type=mask_type,
                                       mask_threshold=mask_threshold)

        self.high_pass_cutoff = 100 if high_pass_cutoff is None else high_pass_cutoff
        self.similarity_threshold = 0 if similarity_threshold is None else similarity_threshold
        self.min_distance_between_frames = 1 if min_distance_between_frames is None else min_distance_between_frames
        self.max_repeating_frames = 100 if max_repeating_frames is None else max_repeating_frames

        self._min_distance_converted_to_hops = False

        self.verbose = False
        self.similarity_matrix = None
        self.background = None
        self.foreground = None
        self.similarity_indices = None
        self.magnitude_spectrogram = None
        self.stft = None
        self.result_masks = None
        self.use_librosa_stft = use_librosa_stft
        self.matlab_fidelity = matlab_fidelity

        if self.matlab_fidelity:
            self.use_librosa_stft = False

        self.do_mono = do_mono

        if self.do_mono:
            self.audio_signal.to_mono(overwrite=True)

    def run(self):
        """
        Runs REPET-SIM, a variant of REPET using the cosine similarity matrix to find similar
        frames to do median filtering.

        Returns:

        """
        # High pass filter cutoff freq. (in # of freq. bins), +1 to match MATLAB implementation
        self.high_pass_cutoff = int(np.ceil(float(self.high_pass_cutoff) *
                                            (self.stft_params.n_fft_bins - 1) /
                                            self.audio_signal.sample_rate) + 1)
        low = 1 if self.matlab_fidelity else 0
        self._compute_spectrograms()
        self.similarity_indices = self._get_similarity_indices()

        background_stft = []
        background_mask = []
        for i in range(self.audio_signal.num_channels):
            repeating_mask = self._compute_mask(self.magnitude_spectrogram[:, :, i])

            repeating_mask[low:self.high_pass_cutoff, :] = 1  # high-pass filter the foreground
            background_mask.append(repeating_mask)

            stft_with_mask = repeating_mask * self.stft[:, :, i]
            background_stft.append(stft_with_mask)

        # Set STFT in correct order
        background_stft = np.array(background_stft).transpose((1, 2, 0))
        self._make_background_signal(background_stft)

        # make a mask and return
        background_mask = np.array(background_mask).transpose((1, 2, 0))
        background_mask = masks.SoftMask(background_mask)
        if self.mask_type == self.BINARY_MASK:
            background_mask = background_mask.mask_to_binary(self.mask_threshold)

        self.result_masks = [background_mask, background_mask.inverse_mask()]

        return self.result_masks

    def _make_background_signal(self, background_stft):
        self.background = self.audio_signal.make_copy_with_stft_data(background_stft, verbose=False)
        self.background.istft(self.stft_params.window_length, self.stft_params.hop_length, self.stft_params.window_type,
                              overwrite=True, use_librosa=self.use_librosa_stft,
                              truncate_to_length=self.audio_signal.signal_length)

    def _compute_spectrograms(self):
        self.stft = self.audio_signal.stft(overwrite=True, remove_reflection=True, use_librosa=self.use_librosa_stft)
        self.magnitude_spectrogram = np.abs(self.stft)

    def _get_similarity_indices(self):
        if self.magnitude_spectrogram is None:
            self._compute_spectrograms()

        self.similarity_matrix = self.get_similarity_matrix()

        if not self._min_distance_converted_to_hops:
            self.min_distance_between_frames *= self.audio_signal.sample_rate / self.stft_params.window_overlap
            self._min_distance_converted_to_hops = True

        return self._find_similarity_indices()

    @staticmethod
    def compute_similarity_matrix(matrix):
        """Computes the cosine similarity matrix using the cosine similarity for any given input matrix X.

        Parameters:
            matrix (np.array): 2D matrix containing the magnitude spectrogram of the audio signal
        Returns:
            S (np.array): 2D similarity matrix
        """
        assert (type(matrix) == np.ndarray)

        # ignore the 'divide by zero' warning
        with np.errstate(divide='ignore'):
            # Code below is adopted from http://stackoverflow.com/a/20687984/5768001

            # base similarity matrix (all dot products)
            similarity = np.dot(matrix, matrix.T)

            # inverse_mask of the squared magnitude of preference vectors (number of occurrences)
            inv_square_mag = 1 / np.diag(similarity)

            # if it doesn't occur, set it's inverse_mask magnitude to zero (instead of inf)
            inv_square_mag[np.isinf(inv_square_mag)] = 0

            # inverse_mask of the magnitude
            inv_mag = np.sqrt(inv_square_mag)

            # cosine similarity (element-wise multiply by inverse_mask magnitudes)
            cosine = similarity * inv_mag
            cosine = cosine.T * inv_mag

            return cosine

    def _find_similarity_indices(self):
        """Finds the similarity indices for all time frames from the similarity matrix

        Returns:
            similarity_indices (list of lists): similarity indices for all time frames
        """
        if self.similarity_matrix is None:
            raise Exception('self.similarity_matrix cannot be None')

        similarity_indices = []
        for i in range(self.audio_signal.stft_length):
            cur_indices = utils.find_peak_indices(self.similarity_matrix[i, :], self.max_repeating_frames,
                                                  min_dist=self.min_distance_between_frames,
                                                  threshold=self.similarity_threshold)

            # the first peak is always itself so we throw it out
            # we also want only self.max_repeating_frames peaks
            # so +1 for 0-based, and +1 for the first peak we threw out
            cur_indices = cur_indices[1:self.max_repeating_frames + 2]
            similarity_indices.append(cur_indices)

        if all(not idx for idx in similarity_indices):
            raise RuntimeError('No similarity indices!')

        if any(not idx for idx in similarity_indices):
            warnings.warn('Not all indices have similarities above threshold!')

        return similarity_indices

    def _compute_mask(self, magnitude_spectrogram_channel):
        """

        Args:
            magnitude_spectrogram_channel:

        Returns:

        """
        if self.magnitude_spectrogram is None:
            self._compute_spectrograms()

        if self.similarity_indices is None:
            self._get_similarity_indices()

        mask = np.zeros_like(magnitude_spectrogram_channel)

        for i in range(self.audio_signal.stft_length):
            cur_similarities = self.similarity_indices[i]

            if not cur_similarities:
                # If there are no similarities, then just add ones to the mask here.
                mask[:, i] = np.ones(mask.shape[constants.STFT_VERT_INDEX])
            else:
                similar_times = np.array([magnitude_spectrogram_channel[:, j]
                                          for j in cur_similarities])

                mask[:, i] = np.median(similar_times, axis=0)

        mask = np.minimum(mask, magnitude_spectrogram_channel)
        mask = (mask + constants.EPSILON) / (magnitude_spectrogram_channel + constants.EPSILON)
        return mask

    def get_similarity_matrix(self):
        """Calculates and returns the similarity matrix for the audio file associated with this object

        Returns:
             similarity_matrix (np.array): similarity matrix for the audio file.

        EXAMPLE::

            # Set up audio signal
            signal = nussl.AudioSignal('path_to_file.wav')

            # Set up a RepetSim object
            repet_sim = nussl.RepetSim(signal)

            # I don't have to run repet to get a similarity matrix for signal
            sim_mat = repet_sim.get_similarity_matrix()

        """
        if self.magnitude_spectrogram is None:
            self._compute_spectrograms()
        mean_magnitude_spectrogram = np.mean(self.magnitude_spectrogram, axis=2)
        return self.compute_similarity_matrix(mean_magnitude_spectrogram.T)

    def make_audio_signals(self):
        """ Returns the background and foreground audio signals. You must have run Repet.run() prior
        to calling this function. This function will return None if run() has not been called.

        Returns:
            Audio Signals (List): 2 element list.

                * bkgd: Audio signal with the calculated background track
                * fkgd: Audio signal with the calculated foreground track

        Example:
            
        .. code-block:: python
            :linenos:
            
            # set up AudioSignal object
            signal = nussl.AudioSignal('path_to_file.wav')

            # set up and run RepetSim
            repet_sim = nussl.RepetSim(signal)
            repet_sim.run()

            # get audio signals (AudioSignal objects)
            background, foreground = repet_sim.make_audio_signals()
        """
        if self.background is None:
            return None

        foreground_array = self.audio_signal.audio_data - self.background.audio_data
        self.foreground = self.audio_signal.make_copy_with_audio_data(foreground_array)
        return [self.background, self.foreground]

    def plot(self, output_file, **kwargs):
        import matplotlib.pyplot as plt
        plt.close('all')

        title = None

        if len(kwargs) != 0:
            if 'title' in kwargs:
                title = kwargs['title']

        sim_mat = self.get_similarity_matrix()
        plt.pcolormesh(sim_mat)
        title = title if title is not None else 'Similarity Matrix for {}'.format(self.audio_signal.file_name)
        plt.xlabel('Time (s)')
        plt.ylabel('Time (s)')
        plt.title(title)

        plt.axis('tight')
        plt.savefig(output_file)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import copy

import numpy as np
import sklearn.cluster
import librosa

from ..transformers import transformer_nmf
import mask_separation_base
import masks


class NMF_MFCC(mask_separation_base.MaskSeparationBase):
    """
        Non Negative Matrix Factorization using K-Means Clustering on MFCC (NMF MFCC) is a source separation
        algorithm that runs Transformer NMF on the magnitude spectrogram of an input audio signal. The templates matrix
        is then converted to mel-space to reduce the dimensionality. The K means clustering then clusters the converted
        templates and activations matrices. The dot product of the clustered templates and activations results in a
        magnitude spectrogram only containing a separated source. This is used to create a Binary Mask object, and the
        whole process can be applied for each cluster to return a list of Audio Signal objects corresponding to each
        separated source.

        References:
            * `Mel Frequency Cepstral Coefficients (MFCC) <https://en.wikipedia.org/wiki/Mel-frequency_cepstrum>`_

            * `scikit-learn KMeans <http://scikit-learn.org/stable/modules/generated/sklearn.cluster.KMeans.html>`_

            * Spiertz, Martin, and Volker Gnann. "Source-filter based clustering for monaural blind source separation."
               Proceedings of the 12th International Conference on Digital Audio Effects. 2009.

        See Also:
            :ref:`The NMF MFCC Demo Example <nmf_mfcc_demo>`

        Parameters:
            input_audio_signal (:class:`audio_signal.AudioSignal`): The :class:`audio_signal.AudioSignal` object that
             NMF MFCC will be run on. This makes a copy of ``input_audio_signal``
            num_sources (int): Number of sources to find.
            num_templates (int): Number of template vectors to used in NMF. Defaults to 50.
            num_iterations (int): The number of iterations to go through in NMF. Defaults to 50.
            random_seed (int): The seed to use in the numpy random generator in NMF and KMeans. See code examples below
             for how this is used. Default uses no seed.
            distance_measure (str): The type of distance measure to use in NMF - euclidean or divergence.
             Defaults to ``euclidean``.
            kmeans_kwargs (dict): The kwargs for KMeans parameters. Can be initialized with a dictionary of keys
             corresponding to parameters in KMeans. See below for an example. Default is ``None``.
            to_mono (bool): Converts signal to mono before running algorithm. Defaults to ``False``.
            mask_type (str): A soft or binary mask object used for the source separation.
             Defaults to :class:`BinaryMask`.
            mfcc_range (int,list,tuple): The range of MFCCs used for clustering. See examples below.
             Defaults to ``1:14``.
            n_mfcc (int): The max number of mfccs to use. Defaults to 20.

        Attributes:
            input_audio_signal (:class:`audio_signal.AudioSignal`): The :class:`audio_signal.AudioSignal` object that
              NMF MFCC will be run on. This makes a copy of ``input_audio_signal``
            clusterer (:obj:`KMeans`): A scikit-learn KMeans object for clustering the templates and activations.
            signal_stft (:obj:`np.ndarray`): The stft data for the input audio signal.
            labeled_templates (:obj:`list`): A Numpy array containing the labeled templates columns
                                               from the templates matrix for a particular source.
            sources (:obj:`list`): A list containing the lists of Audio Signal objects for each source.
            result_masks (:obj:`list`): A list containing the lists of Binary Mask objects for each channel.

         Initializing Example:

        .. code-block:: python
            :linenos:
            # Initialize input signal
            signal = nussl.AudioSignal(path_to_input_file='input_name.wav')

            # Default initialization
             nmf_mfcc =  nussl.NMF_MFCC(signal, num_sources=2)

            # Random Seeding initialization
            # Set NMF and KMeans seeds to 0 (no kmeans_kwarg initialized, so random_seed is used for random_state)
            nmf_mfcc =  nussl.NMF_MFCC(signal, num_sources=2, random_seed=0)

            # Use individual seeding
            nmf_mfcc =  nussl.NMF_MFCC(signal, num_sources=2, random_seed=0, kmeans_kwargs = {random_state: 1})

            # Only set random_seed and not random_state
            nmf_mfcc =  nussl.NMF_MFCC(signal, num_sources=2, random_seed=0, kmeans_kwargs = {random_state: None})

            # MFCC range initialization
            # Set up the max of the MFCC range by only using an int
            nmf_mfcc = nussl.NMF_MFCC(signal, num_sources=2, num_templates=6, mfcc_range=5)

            # Set up the MFCC range by using a list [min, max]
            nmf_mfcc = nussl.NMF_MFCC(signal, num_sources=2, num_templates=6, mfcc_range=[3, 15])

            # Set up the MFCC range by using a tuple (min, max)
            nmf_mfcc = nussl.NMF_MFCC(signal, num_sources=2, num_templates=6, mfcc_range=(2, 14))

            # KMeans initialization
            # Initialize all KMeans arguments, see KMeans documentation for more detail.
            kmeans_kwargs = { n_clusters: 8, init: 'k-means++', n_init: 10, max_iter: 300, tol: 1e-4,
                              precompute_distances: 'auto', verbose: 0, random_state: None, copy_x: True,
                              n_jobs: 1, algorithm: 'auto'}
            nmf_mfcc =  nussl.NMF_MFCC(signal, num_sources=2, kmeans_kwargs=kmeans_kwargs)

        """
    def __init__(self, input_audio_signal, num_sources, num_templates=50, num_iterations=50, random_seed=None,
                 distance_measure=transformer_nmf.TransformerNMF.EUCLIDEAN, kmeans_kwargs=None, to_mono=False,
                 mask_type=mask_separation_base.MaskSeparationBase.BINARY_MASK, mfcc_range=(1, 14), n_mfcc=20):
        super(NMF_MFCC, self).__init__(input_audio_signal=input_audio_signal, mask_type=mask_type)

        self.num_sources = num_sources
        self.num_templates = num_templates
        self.distance_measure = distance_measure
        self.num_iterations = num_iterations
        self.random_seed = random_seed
        self.kmeans_kwargs = kmeans_kwargs
        self.n_mfcc = n_mfcc

        self.signal_stft = None
        self.labeled_templates = None
        self.sources = []

        # Convert the stereo signal to mono if indicated
        if to_mono:
            self.audio_signal.to_mono(overwrite=True)

        # Set the MFCC range
        if isinstance(mfcc_range, int) and mfcc_range < n_mfcc:
            self.mfcc_start, self.mfcc_end = 1, mfcc_range
        elif isinstance(mfcc_range, (tuple, list)) and len(mfcc_range) == 2:
            self.mfcc_start, self.mfcc_end = mfcc_range[0], mfcc_range[1]
        else:
            raise ValueError('mfcc_range is not set correctly! Must be a tuple or list with (min, max), or int (max)')

        # If kmeans_kwargs does not include the 'random_state', use the random_seed instead. Else, use the value
        # provided in the dictionary. If kmeans_kwargs is None, use the random_seed.
        self.kmeans_random_seed = kmeans_kwargs.pop('random_state', random_seed) \
            if isinstance(kmeans_kwargs, dict) else random_seed

        # Initialize the K Means clusterer
        if isinstance(self.kmeans_kwargs, dict):
            self.clusterer = sklearn.cluster.KMeans(n_clusters=self.num_sources, random_state=self.kmeans_random_seed,
                                                    **kmeans_kwargs)
        else:
            self.clusterer = sklearn.cluster.KMeans(n_clusters=self.num_sources, random_state=self.kmeans_random_seed)

    def run(self):
        """ This function calls TransformerNMF on the magnitude spectrogram of each channel in the input audio signal.
        The templates and activation matrices returned are clustered using K-Means clustering. These clusters are used
        to create mask objects for each source. Note: The masks in self.result_masks are not returned in a particular
        order corresponding to the sources, but they are in the same order for each channel.

        Returns:
            result_masks (list): A list of :obj:`MaskBase`-derived objects for each source.
            (to get a list of :obj:`AudioSignal`-derived objects run :func:`make_audio_signals`)

        Example:

        .. code-block:: python
            :linenos:

            signal = nussl.AudioSignal(path_to_input_file='input_name.wav')

            # Set up and run NMF MFCC
            nmf_mfcc =  nussl.NMF_MFCC(signal, num_sources=2) # Returns a binary mask by default
            masks = nmf_mfcc.run()

            # Get audio signals
            sources = nmf_mfcc.make_audio_signals()

            # Output the sources
            for i, source in enumerate(sources):
                output_file_name = str(i) + '.wav'
                source.write_audio_to_file(output_file_name)
        """
        self.audio_signal.stft_params = self.stft_params
        self.audio_signal.stft()

        uncollated_masks = []
        n_chan = self.audio_signal.num_channels
        for ch in range(n_chan):
            channel_stft = self.audio_signal.get_magnitude_spectrogram_channel(ch)

            # Set up NMF and run
            nmf = transformer_nmf.TransformerNMF(input_matrix=channel_stft, num_components=self.num_templates,
                                                 seed=self.random_seed, should_do_epsilon=False,
                                                 max_num_iterations=self.num_iterations,
                                                 distance_measure=self.distance_measure)

            channel_activation_matrix, channel_templates_matrix = nmf.transform()

            # Cluster the templates matrix into Mel frequencies and retrieve labels
            cluster_templates = librosa.feature.mfcc(S=channel_templates_matrix,
                                                     n_mfcc=self.n_mfcc)[self.mfcc_start:self.mfcc_end]
            self.clusterer.fit_transform(cluster_templates.T)
            self.labeled_templates = self.clusterer.labels_

            # Extract sources from signal
            uncollated_masks += self._extract_masks(channel_templates_matrix, channel_activation_matrix, ch)

        # Reorder mask arrays so that the channels are collated correctly (this allows for multichannel signals)
        collated_masks = [np.dstack([uncollated_masks[s + ch * self.num_sources] for ch in range(n_chan)])
                          for s in range(self.num_sources)]

        # Put each numpy array mask into a MaskBase object
        self.result_masks = []
        for mask in collated_masks:
            if self.mask_type == self.BINARY_MASK:
                mask = np.round(mask)
                mask_object = masks.BinaryMask(mask)
            elif self.mask_type == self.SOFT_MASK:
                mask_object = masks.SoftMask(mask)
            else:
                raise ValueError('Unknown mask type {}!'.format(self.mask_type))
            self.result_masks.append(mask_object)

        return self.result_masks

    def _extract_masks(self, templates_matrix, activation_matrix, ch):
        """ Creates binary masks from clustered templates and activation matrices

        Parameters:
            templates_matrix (np.ndarray): A 2D Numpy array containing the templates matrix after running NMF on
                                          the current channel
            activation_matrix (np.ndarray): A 2D Numpy array containing the activation matrix after running NMF on
                                          the current channel

        Returns:
            channel_mask_list (list): A list of :obj:`MaskBase`-derived objects corresponding to each source
        """

        if self.audio_signal.stft_data is None:
            raise ValueError('Cannot extract masks with no signal_stft data')

        channel_mask_list = []
        for source_index in range(self.num_sources):
            source_indices = np.where(self.labeled_templates == source_index)[0]

            # Make empty copies of the templates and actications
            templates_mask = np.zeros_like(templates_matrix)
            activation_mask = np.zeros_like(activation_matrix)

            # Keep values for each source as determined by the clusterer
            for idx in source_indices:
                templates_mask[:, idx] = templates_matrix[:, idx]
                activation_mask[idx, :] = activation_matrix[idx, :]

            mask_matrix = templates_mask.dot(activation_mask)
            music_stft_max = np.maximum(mask_matrix, np.abs(self.audio_signal.get_stft_channel(ch)))
            mask_matrix = np.divide(mask_matrix, music_stft_max)
            mask = np.nan_to_num(mask_matrix)

            channel_mask_list.append(mask)

        return channel_mask_list

    def make_audio_signals(self):
        """ Applies each mask in self.masks and returns a list of audio_signal objects for each source.

        Returns:
            self.sources (np.array): An array of audio_signal objects containing each separated source
        """
        self.sources = []
        for mask in self.result_masks:

            source = copy.deepcopy(self.audio_signal)
            source = source.apply_mask(mask)
            source.stft_params = self.stft_params
            source.istft(overwrite=True, truncate_to_length=self.audio_signal.signal_length)
            self.sources.append(source)

        return self.sources



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Base class for separation algorithms that make masks. Most algorithms in nussl are derived from MaskSeparationBase. 

"""
import json
import warnings

import masks
import separation_base
from ..core import utils
from ..core import audio_signal
from ..core import constants


class MaskSeparationBase(separation_base.SeparationBase):
    """
    Base class for separation algorithms that create a mask (binary or soft) to do their separation. Most algorithms 
    in nussl are derived from :class:`MaskSeparationBase`.
    
    Although this class will do nothing if you instantiate and run it by itself, algorithms that are derived from this
    class are expected to return a list of :class:`separation.masks.mask_base.MaskBase` -derived objects 
    (i.e., either a :class:`separation.masks.binary_mask.BinaryMask` or :class:`separation.masks.soft_mask.SoftMask`
    object) by their :func:`run()` method. Being a subclass of :class:`MaskSeparationBase` is an implicit contract 
    assuring this.  Returning a :class:`separation.masks.mask_base.MaskBase`-derived object standardizes 
    algorithm return types for :class:`evaluation.evaluation_base.EvaluationBase`-derived objects. 
    
    Args:
        input_audio_signal: (:class:`audio_signal.AudioSignal`) An :class:`audio_signal.AudioSignal` object containing 
            the mixture to be separated.
        mask_type: (str) Indicates whether to make binary or soft masks. See :attr:`mask_type` property for details.
        mask_threshold: (float) Value between [0.0, 1.0] to convert a soft mask to a binary mask. See 
            :attr:`mask_threshold` property for details.
    
    """

    BINARY_MASK = 'binary'
    """ String alias for setting this object to return :class:`separation.masks.binary_mask.BinaryMask` objects
    """

    SOFT_MASK = 'soft'
    """ String alias for setting this object to return :class:`separation.masks.soft_mask.SoftMask` objects
    """

    _valid_mask_types = [BINARY_MASK, SOFT_MASK]

    def __init__(self, input_audio_signal, mask_type=SOFT_MASK, mask_threshold=0.5):
        super(MaskSeparationBase, self).__init__(input_audio_signal=input_audio_signal)

        self._mask_type = None
        self.mask_type = mask_type
        self._mask_threshold = None
        self.mask_threshold = mask_threshold
        self.result_masks = []

    @property
    def mask_type(self):
        """
        PROPERTY
        
        This property indicates what type of mask the derived algorithm will create and be returned by :func:`run()`.
        Options are either 'soft' or 'binary'. 
        :attr:`mask_type` is usually set when initializing a :class:`MaskSeparationBase`-derived class 
        and defaults to :attr:`SOFT_MASK`.
        
        This property, though stored as a string, can be set in two ways when initializing:
        
        * First, it is possible to set this property with a string. Only ``'soft'`` and ``'binary'`` are accepted 
          (case insensitive), every other value will raise an error. When initializing with a string, two helper 
          attributes are provided: :attr:`BINARY_MASK` and :attr:`SOFT_MASK`.
        
          It is **HIGHLY** encouraged to use these, as the API may change and code that uses bare strings 
          (e.g. ``mask_type = 'soft'`` or ``mask_type = 'binary'``) for assignment might not be future-proof. 
          :attr:`BINARY_MASK`` and :attr:`SOFT_MASK` are safe aliases in case these underlying types change.
        
        * The second way to set this property is by using a class prototype of either the 
          :class:`separation.masks.binary_mask.BinaryMask` or :class:`separation.masks.soft_mask.SoftMask` class
          prototype. This is probably the most stable way to set this, and it's fairly succinct. 
          For example, ``mask_type = nussl.BinaryMask`` or ``mask_type = nussl.SoftMask`` are both perfectly valid.
        
        Though uncommon, this can be set outside of :func:`__init__()` 
        
        Examples of both methods are shown below.
        
        Returns:
            mask_type (str): Either ``'soft'`` or ``'binary'``. 
            
        Raises:
            ValueError if set invalidly.
            
        Example:
            
        .. code-block:: python
            :linenos:
    
            import nussl
            mixture_signal = nussl.AudioSignal()
                
            # Two options for determining mask upon init...
             
            # Option 1: Init with a string (BINARY_MASK is a string 'constant')
            repet_sim = nussl.RepetSim(mixture_signal, mask_type=nussl.MaskSeparationBase.BINARY_MASK)
             
            # Option 2: Init with a class type
            ola = nussl.OverlapAdd(mixture_signal, mask_type=nussl.SoftMask)
             
            # It's also possible to change these values after init by changing the `mask_type` property...
            repet_sim.mask_type = nussl.MaskSeparationBase.SOFT_MASK  # using a string
            ola.mask_type = nussl.BinaryMask  # or using a class type
            
        """
        return self._mask_type

    @mask_type.setter
    def mask_type(self, value):
        error = ValueError('Invalid mask type! Got {0} but valid masks are: [{1}]!'
                           .format(value, ', '.join(self._valid_mask_types)))

        if value is None:
            raise error

        if isinstance(value, str):
            value = value.lower()
            if value in self._valid_mask_types:
                self._mask_type = value
            else:
                raise error

        elif isinstance(value, masks.MaskBase):
            warnings.warn('This separation method is not using the values in the provided mask object.')
            value = type(value).__name__
            value = value[:value.find('Mask')].lower()

            if value not in self._valid_mask_types:
                # make sure we don't get duped by accident. This shouldn't happen
                raise error
            self._mask_type = value

        elif issubclass(value, masks.MaskBase):
            if value is masks.BinaryMask:
                self._mask_type = self.BINARY_MASK
            elif value is masks.SoftMask:
                self._mask_type = self.SOFT_MASK
            else:
                raise error

        else:
            raise error

    @property
    def mask_threshold(self):
        """
        PROPERTY
        
        Threshold of determining True/False if :attr:`mask_type` is :attr:`BINARY_MASK`. Some algorithms will first 
        make a soft mask and then convert that to a binary mask using this threshold parameter. All values of the 
        soft mask are between ``[0.0, 1.0]`` and as such :func:`mask_threshold` is expected to be a float between 
        ``[0.0, 1.0]``.
        
        Returns:
            mask_threshold (float): Value between ``[0.0, 1.0]`` that indicates the True/False cutoff when converting a
            soft mask to binary mask.
                                
        Raises:
            ValueError if not a float or if set outside ``[0.0, 1.0]``.

        """
        return self._mask_threshold

    @mask_threshold.setter
    def mask_threshold(self, value):
        if not isinstance(value, float) or not (0.0 < value < 1.0):
            raise ValueError('Mask threshold must be a float between [0.0, 1.0]!')

        self._mask_threshold = value

    def zeros_mask(self, shape):
        """
        Creates a new zeros mask with this object's type

        Args:
            shape:

        Returns:

        """
        if self.mask_type == self.BINARY_MASK:
            return masks.BinaryMask.zeros(shape)
        else:
            return masks.SoftMask.zeros(shape)

    def ones_mask(self, shape):
        """

        Args:
            shape:

        Returns:

        """
        if self.mask_type == self.BINARY_MASK:
            return masks.BinaryMask.ones(shape)
        else:
            return masks.SoftMask.ones(shape)

    def plot(self, output_name, **kwargs):
        """Plots relevant data for mask-based separation algorithm. Base class: Do not call directly!

        Raises:
            NotImplementedError: Cannot call base class!
        """
        raise NotImplementedError('Cannot call base class!')

    def run(self):
        """Runs mask-based separation algorithm. Base class: Do not call directly!

        Raises:
            NotImplementedError: Cannot call base class!
        """
        raise NotImplementedError('Cannot call base class!')

    def make_audio_signals(self):
        """Makes :class:`audio_signal.AudioSignal` objects after mask-based separation algorithm is run. 
        Base class: Do not call directly!

        Raises:
            NotImplementedError: Cannot call base class!
        """
        raise NotImplementedError('Cannot call base class!')

    @classmethod
    def from_json(cls, json_string):
        """
        Creates a new :class:`SeparationBase` object from the parameters stored in this JSON string.

        Args:
            json_string (str): A JSON string containing all the data to create a new :class:`SeparationBase`
                object.

        Returns:
            (:class:`SeparationBase`) A new :class:`SeparationBase` object from the JSON string.

        See Also:
            :func:`to_json` to make a JSON string to freeze this object.

        """
        mask_sep_decoder = MaskSeparationBaseDecoder(cls)
        return mask_sep_decoder.decode(json_string)


class MaskSeparationBaseDecoder(separation_base.SeparationBaseDecoder):
    """ Object to decode a :class:`MaskSeparationBase`-derived object from JSON serialization.
    You should never have to instantiate this object by hand.
    """

    def __init__(self, separation_class):
        self.separation_class = separation_class
        json.JSONDecoder.__init__(self, object_hook=self._json_separation_decoder)

    def _json_separation_decoder(self, json_dict):
        if '__class__' in json_dict and '__module__' in json_dict:
            json_dict, separator = self._inspect_json_and_create_new_instance(json_dict)

            # fill out the rest of the fields
            for k, v in json_dict.items():
                if isinstance(v, dict) and constants.NUMPY_JSON_KEY in v:
                    separator.__dict__[k] = utils.json_numpy_obj_hook(v[constants.NUMPY_JSON_KEY])
                    
                # TODO: test this in python3
                elif isinstance(v, (str, bytes, unicode)) and audio_signal.__name__ in v:

                    separator.__dict__[k] = audio_signal.AudioSignal.from_json(v)
                elif k == 'result_masks':
                    # for mask_json in v:

                    separator.result_masks = [masks.MaskBase.from_json(itm) for itm in v]
                else:
                    separator.__dict__[k] = v if not isinstance(v, unicode) else v.encode('ascii')

            return separator
        else:
            return json_dict



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Base class for Mask objects. Contains many common utilities used for accessing masks. The mask itself is
represented under the hood as a three dimensional numpy :obj:`ndarray` object. The dimensions are 
``[NUM_FREQ, NUM_HOPS, NUM_CHAN]``. Safe accessors for these array indices are in :ref:`constants` as well as 
below.

Right now only spectrogram-like masks are supported (note the shape of the :ref:`mask` property), but in future
releases nussl will support masks for representations with different dimensionality requirements.
"""

import copy
import json
import numbers

import numpy as np

from ...core import utils
from ...core import constants


class MaskBase(object):
    """
    Args:
        input_mask (:obj:`np.ndarray`): A 2- or 3-dimensional numpy ``ndarray`` representing a mask.
        
    """
    def __init__(self, input_mask=None, mask_shape=None):
        self._mask = None

        if mask_shape is None and input_mask is None:
            raise ValueError('Cannot initialize mask without mask_shape or input_mask!')

        if mask_shape is not None and input_mask is not None:
            raise ValueError('Cannot initialize mask with both mask_shape and input_mask!')

        if isinstance(input_mask, np.ndarray):
            self.mask = input_mask

        elif isinstance(mask_shape, tuple):
            self.mask = np.zeros(mask_shape)

        else:
            raise ValueError('input_mask must be a np.ndarray, or mask_shape must be a tuple!')

    @property
    def mask(self):
        """
        PROPERTY
        
        The actual mask. This is represented as a three dimensional numpy :obj:`ndarray` object.
        The input gets validated by :func:`_validate_mask`. In the case of 
        :class:`separation.masks.binary_mask.BinaryMask` the validation checks that the values are all 1 or 0 
        (or bools), in the case of :class:`separation.masks.soft_mask.SoftMask` the validation checks that all values
        are within the domain ``[0.0, 1.0]``.
        
        This base class will throw a ``NotImplementedError`` if instantiated directly.
        
        Raises:
            :obj:`ValueError` if :attr:`mask.ndim` is less than 2 or greater than 3, or if values fail validation.
            :obj:`NotImplementedError` if instantiated directly.

        """
        return self._mask

    @mask.setter
    def mask(self, value):
        assert isinstance(value, np.ndarray), 'Type of self.mask must be np.ndarray!'

        if value.ndim == 1:
            raise ValueError('Cannot support arrays with less than 2 dimensions!')

        if value.ndim == 2:
            value = np.expand_dims(value, axis=constants.STFT_CHAN_INDEX)

        if value.ndim > 3:
            raise ValueError('Cannot support arrays with more than 3 dimensions!')

        self._mask = self._validate_mask(value)

    def get_channel(self, n):
        """
        Gets mask channel ``n`` and returns it as a 2D :obj:`np.ndarray`
        
        Args:
            n (int): Channel index to return (0-based).

        Returns:
            :obj:`np.array` with the mask channel
            
        Raises:
            :obj:`AttributeError` if :attr:`mask` is ``None``
            :obj:`ValueError` if ``n`` is less than 0 or greater than the number of channels that this mask object has.

        """
        if self.mask is None:
            raise AttributeError('Cannot get channel {} when mask has no data!'.format(n))

        if n >= self.num_channels:
            raise ValueError('Cannot get channel {0} when this object only has {1} channels! (0-based)'
                             .format(n, self.num_channels))

        if n < 0:
            raise ValueError('Cannot get channel {}. This will cause unexpected results!'.format(n))

        return utils._get_axis(self.mask, constants.STFT_CHAN_INDEX, n)

    @property
    def length(self):
        """
        (int) Number of time hops that this mask represents.

        """
        if self.mask is None:
            raise AttributeError('Cannot get length of BinaryMask when there is no mask data!')
        return self.mask.shape[constants.STFT_LEN_INDEX]

    @property
    def height(self):
        """
        (int) Number of frequency bins this mask has.

        """
        if self.mask is None:
            raise AttributeError('Cannot get height of BinaryMask when there is no mask data!')
        return self.mask.shape[constants.STFT_VERT_INDEX]

    @property
    def num_channels(self):
        """
        (int) Number of channels this mask has.

        """
        if self.mask is None:
            raise AttributeError('Cannot get num_channels of BinaryMask when there is no mask data!')
        return self.mask.shape[constants.STFT_CHAN_INDEX]

    @property
    def shape(self):
        """
        (tuple) Returns the shape of the whole mask. Identical to ``np.ndarray.shape()``.

        """
        if self.mask is None:
            raise AttributeError('Cannot get shape of BinaryMask when there is no mask data!')
        return self.mask.shape

    @property
    def dtype(self):
        """
        (str) Returns the data type of the values of the mask. 

        """
        if self.mask is None:
            raise AttributeError('Cannot get num_channels of BinaryMask when there is no mask data!')
        return self.mask.dtype

    @staticmethod
    def _validate_mask(mask_):
        """
        Args:
            mask_: 

        Returns:

        """
        raise NotImplementedError('Cannot call base class! Use BinaryMask or SoftMask!')

    @classmethod
    def ones(cls, shape):
        """
        Makes a mask with all ones with the specified shape. Exactly the same as ``np.ones()``.
        Args:
            shape (tuple): Shape of the resultant mask.

        Returns:

        """
        return cls(np.ones(shape))

    @classmethod
    def zeros(cls, shape):
        """
        Makes a mask with all zeros with the specified shape. Exactly the same as ``np.zeros()``.
        Args:
            shape (tuple): Shape of the resultant mask.

        Returns:

        """
        return cls(np.zeros(shape))

    def invert_mask(self):
        """

        Returns:

        """
        raise NotImplementedError('Cannot call base class! Use BinaryMask or SoftMask!')

    def inverse_mask(self):
        """
        Alias for :func:`invert_mask`

        See Also:
            :func:`invert_mask`

        Returns:

        """
        return self.invert_mask()

    def _add(self, other):
        if isinstance(other, MaskBase):
            return self.mask + other.mask
        if isinstance(other, np.ndarray):
            return self.mask + other

        else:
            raise ValueError('Cannot do arithmetic operation with MaskBase and {}'.format(type(other)))

    def _mult(self, value):
        if not isinstance(value, numbers.Real):
            raise ValueError('Cannot do operation with MaskBase and {}'.format(type(value)))

        return self.mask * value

    def to_json(self):
        """

        Returns:

        """
        return json.dumps(self, default=MaskBase._to_json_helper)

    @staticmethod
    def _to_json_helper(o):
        if not isinstance(o, MaskBase):
            raise TypeError('MaskBase._to_json_helper() got foreign object!')

        d = copy.copy(o.__dict__)
        for k, v in d.items():
            if isinstance(v, np.ndarray):
                d[k] = utils.json_ready_numpy_array(v)

        d['__class__'] = o.__class__.__name__
        d['__module__'] = o.__module__
        if 'self' in d:
            del d['self']

        return d

    @classmethod
    def from_json(cls, json_string):
        """ Creates a new :class:`MaskBase` object from the parameters stored in this JSON string.

        Args:
            json_string (str): A JSON string containing all the data to create a new :class:`MaskBase`
                object.

        Returns:
            (:class:`SeparationBase`) A new :class:`MaskBase` object from the JSON string.

        See Also:
            :func:`to_json` to make a JSON string to freeze this object.

        """
        mask_decoder = MaskBaseDecoder(cls)
        return mask_decoder.decode(json_string)

    def __add__(self, other):
        return self._add(other)

    def __sub__(self, other):
        return self + (-1 * other)

    def __iadd__(self, other):
        return self + other

    def __isub__(self, other):
        return self - other

    def __mul__(self, value):
        return self._mult(value)

    def __div__(self, value):
        return self._mult(1 / float(value))

    def __truediv__(self, value):
        return self.__div__(value)

    def __imul__(self, value):
        return self * value

    def __idiv__(self, value):
        return self / value

    def __itruediv__(self, value):
        return self.__idiv__(value)

    def __eq__(self, other):
        return np.array_equal(self.mask, other.mask)

    def __ne__(self, other):
        return not self.__eq__(other)


class MaskBaseDecoder(json.JSONDecoder):
    """ Object to decode a :class:`MaskBase`-derived object from JSON serialization.
    You should never have to instantiate this object by hand.
    """

    def __init__(self, mask_class):
        self.mask_class = mask_class
        json.JSONDecoder.__init__(self, object_hook=self._json_mask_decoder)

    def _json_mask_decoder(self, json_dict):
        """
        Helper method for :class:`MaskBaseDecoder`. Don't you worry your pretty little head about this.

        NEVER CALL THIS DIRECTLY!!

        Args:
            json_dict (dict): JSON dictionary provided by `object_hook`

        Returns:
            A new :class:`MaskBase`-derived object from JSON serialization

        """
        if '__class__' in json_dict and '__module__' in json_dict:
            class_name = json_dict.pop('__class__')
            module_name = json_dict.pop('__module__')

            mask_modules, mask_names = zip(*[(c.__module__, c.__name__) for c in MaskBase.__subclasses__()])

            if class_name not in mask_names or module_name not in mask_modules:
                raise TypeError('Got unknown mask type ({}.{}) from json!'.format(module_name, class_name))

            # load the module and import the class
            module = __import__(module_name).separation.masks
            class_ = getattr(module, class_name)

            if '_mask' not in json_dict:
                raise TypeError('JSON string from {} does not have mask!'.format(class_name))

            mask_json = json_dict.pop('_mask')  # this is the mask numpy array
            mask_numpy = utils.json_numpy_obj_hook(mask_json[constants.NUMPY_JSON_KEY])

            return class_(input_mask=mask_numpy)
        else:
            return json_dict



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import division

import json
import os.path
import warnings

import librosa
import matplotlib.pyplot as plt
import numpy as np
import scipy.fftpack as scifft
import scipy.signal

import constants

__all__ = ['plot_stft', 'e_stft', 'e_istft', 'e_stft_plus', 'librosa_stft_wrapper', 'librosa_istft_wrapper',
           'make_window', 'StftParams']


def plot_stft(signal, file_name, title=None, win_length=None, hop_length=None,
              window_type=None, sample_rate=constants.DEFAULT_SAMPLE_RATE, n_fft_bins=None,
              freq_max=None, show_interactive_plot=False):
    """
    Outputs an image of an stft plot of input audio, signal. This uses matplotlib to create the output file.
    You can specify the same all of the same parameters that are in e_stft(). By default, the StftParams defaults
    are used for any values not provided in (win_length, hop_length, and window_type).
    Title is settable by user and there is a flag to show an interactive matplotlib graph, as well.

    Notes:
        To find out what output formats are available for your machine run the following code:

        >>> import matplotlib.pyplot as plt
        >>> fig = plt.figure()
        >>> print(fig.canvas.get_supported_filetypes())
    
        (From here: http://stackoverflow.com/a/7608273/5768001)

    Args:
        signal: (np.array) input time series signal that will be plotted
        file_name: (str) path to file that will be output. Will overwrite any file that is already there.
            Uses mat
        title: (string) (Optional) Title to go at top of graph. Defaults to 'Spectrogram of [file_name]'
        win_length: (int) (Optional) number of samples per window. Defaults to StftParams default.
        hop_length: (int) (Optional) number of samples between the start of adjacent windows, or "hop".
            Defaults to StftParams default.
        sample_rate: (int) (Optional) sample rate of input signal.  Defaults to StftParams default.
        window_type: (string) (Optional) type of window to use. Using WindowType object is recommended.
            Defaults to StftParams default.
        n_fft_bins: (int) (Optional) number of fft bins per time window.
            If not specified, defaults to next highest power of 2 above window_length. Defaults to StftParams default.
        freq_max: (int) Max frequency to display. Defaults to 44100Hz
        show_interactive_plot: (bool) (Optional) Flag indicating if plot should be shown when function is run.
            Defaults to False

    Example:
        
    .. code-block:: python
        :linenos:
        
        # Set up sine wave parameters
        sr = nussl.Constants.DEFAULT_SAMPLE_RATE # 44.1kHz
        n_sec = 3 # seconds
        duration = n_sec * sr
        freq = 300 # Hz
    
        # Make sine wave array
        x = np.linspace(0, freq * 2 * np.pi, duration)
        x = np.sin(x)
    
        # plot it and save it in path 'path/to/sine_wav.png'
        nussl.plot_stft(x, 'path/to/sine_wav.png')

    """
    freq_max = freq_max if freq_max is not None else sample_rate // 2

    if title is None:
        title = os.path.basename(file_name)
        title = os.path.splitext(title)[0]
        title = 'Spectrogram of {}'.format(title)

    required = [win_length, hop_length, window_type]
    if any([r is None for r in required]):
        defaults = StftParams(sample_rate)

        win_length = defaults.window_length if win_length is None else win_length
        hop_length = defaults.hop_length if hop_length is None else hop_length
        window_type = defaults.window_type if window_type is None else window_type

    (stft, psd, freqs, time) = e_stft_plus(signal, win_length, hop_length,
                                           window_type, sample_rate, n_fft_bins)

    plt.close('all')

    # TODO: remove transposes!
    time_tile = np.tile(time, (len(freqs), 1))
    freq_tile = np.tile(freqs.T, (len(time), 1)).T
    sp = librosa.logamplitude(np.abs(stft) ** 2, ref_power=np.max)
    plt.pcolormesh(time_tile, freq_tile, sp)

    plt.axis('tight')
    plt.xlabel('Time (sec)')
    plt.ylabel('Frequency (Hz)')
    plt.title(title)
    plt.ylim(freqs[0], freq_max)

    plt.savefig(file_name)

    if show_interactive_plot:
        plt.interactive('True')
        plt.show()


def e_stft(signal, window_length, hop_length, window_type,
           n_fft_bins=None, remove_reflection=True, remove_padding=False):
    """
    This function computes a short time fourier transform (STFT) of a 1D numpy array input signal.
    This will zero pad the signal by half a hop_length at the beginning to reduce the window
    tapering effect from the first window. It also will zero pad at the end to get an integer number of hops.

    By default, this function removes the FFT data that is a reflection from over Nyquist. There is an option
    to suppress this behavior and have this function include data from above Nyquist, but since the
    inverse STFT function, e_istft(), expects data without the reflection, the onus is on the user to remember
    to set the reconstruct_reflection flag in e_istft() input.

    Additionally, this function assumes a single channeled audio signal and is not guaranteed to work on
    multichannel audio. If you want to do an STFT on multichannel audio see the AudioSignal object.

    Args:
        signal: 1D numpy array containing audio data. (REAL)
        window_length: (int) number of samples per window
        hop_length: (int) number of samples between the start of adjacent windows, or "hop"
        window_type: (string) type of window to use. Using WindowType object is recommended.
        n_fft_bins: (int) (Optional) number of fft bins per time window.
        If not specified, defaults to next highest power of 2 above window_length
        remove_reflection: (bool) (Optional) if True, this will remove reflected STFT data above the Nyquist point.
        If not specified, defaults to True.
        remove_padding: (bool) (Optional) if True, this will remove the extra padding added when doing the STFT.
        Defaults to True.

    Returns:
        2D  numpy array with complex STFT data.
        Data is of shape (num_time_blocks, num_fft_bins). These numbers are determined by length of the input signal,
        on internal zero padding (explained at top), and n_fft_bins/remove_reflection input (see example below).

    Example:
        
    .. code-block:: python
        :linenos:
        
        
        # Set up sine wave parameters
        sr = nussl.Constants.DEFAULT_SAMPLE_RATE # 44.1kHz
        n_sec = 3 # seconds
        duration = n_sec * sr
        freq = 300 # Hz

        # Make sine wave array
        x = np.linspace(0, freq * 2 * np.pi, duration)
        x = np.sin(x)

        # Set up e_stft() parameters
        win_type = nussl.WindowType.HANN
        win_length = 2048
        hop_length = win_length / 2

        # Run e_stft()
        stft = nussl.e_stft(x, win_length, hop_length, win_type)
        # stft has shape (win_length // 2 + 1 , duration / hop_length)

        # Get reflection
        stft_with_reflection = nussl.e_stft(x, win_length, hop_length, win_type, remove_reflection=False)
        # stft_with_reflection has shape (win_length, duration / hop_length)

        # Change number of fft bins per hop
        num_bins = 4096
        stft_more_bins = e_stft(x, win_length, hop_length, win_type, n_fft_bins=num_bins)
        # stft_more_bins has shape (num_bins // 2 + 1, duration / hop_length)
        
    """
    if n_fft_bins is None:
        n_fft_bins = window_length

    window_type = constants.WINDOW_DEFAULT if window_type is None else window_type
    window = make_window(window_type, window_length)

    orig_signal_length = len(signal)
    signal, num_blocks = _add_zero_padding(signal, window_length, hop_length)
    sig_zero_pad = len(signal)
    # figure out size of output stft
    stft_bins = n_fft_bins // 2 + 1 if remove_reflection else n_fft_bins  # only want just over half of each fft

    # this is where we do the stft calculation
    stft = np.zeros((num_blocks, stft_bins), dtype=complex)
    for hop in range(num_blocks):
        start = hop * hop_length
        end = start + window_length
        unwindowed_signal = signal[start:end]
        windowed_signal = np.multiply(unwindowed_signal, window)
        fft = scifft.fft(windowed_signal, n=n_fft_bins)
        stft[hop, ] = fft[:stft_bins]

    # reshape the 2d array, so it's (n_fft, n_hops).
    stft = stft.T
    stft = _remove_stft_padding(stft, orig_signal_length, window_length, hop_length) if remove_padding else stft

    return stft

def librosa_stft_wrapper(signal, window_length, hop_length, window_type=None, remove_reflection=True,
                         center=True, n_fft_bins=None):
    """

    Args:
        signal:
        window_length:
        hop_length:
        window_type:
        remove_reflection:
        center:
        n_fft_bins:

    Returns:

    """

    if window_type is not None and n_fft_bins is not None:
        warnings.warn("n_fft_bins ignored. Librosa's stft uses window_length as n_fft_bins")

    window = make_window(window_type, window_length) if window_type is not None else None
    signal = librosa.util.fix_length(signal, len(signal) + hop_length)
    stft = librosa.stft(signal, n_fft=window_length, hop_length=hop_length, win_length=window_length,
                        window=window, center=center)

    stft = stft if remove_reflection else _add_reflection(stft)

    return stft


def e_istft(stft, window_length, hop_length, window_type, reconstruct_reflection=True, remove_padding=True):
    """
    Computes an inverse_mask short time fourier transform (STFT) from a 2D numpy array of complex values. By default
    this function assumes input STFT has no reflection above Nyquist and will rebuild it, but the
    reconstruct_reflection flag overrides that behavior.

    Additionally, this function assumes a single channeled audio signal and is not guaranteed to work on
    multichannel audio. If you want to do an iSTFT on multichannel audio see the AudioSignal object.

    Args:
        stft: complex valued 2D numpy array containing STFT data
        window_length: (int) number of samples per window
        hop_length: (int) number of samples between the start of adjacent windows, or "hop"
        window_type: (deprecated)
        reconstruct_reflection: (bool) (Optional) if True, this will recreate the removed reflection
        data above the Nyquist. If False, this assumes that the input STFT is complete. Default is True.
        remove_padding: (bool) (Optional) if True, this function will remove the first and
            last (window_length - hop_length) number of samples. Defaults to False.
        will massage the output so that it is in a format that it expects. remove_reflection is still works in this
        mode. Note: librosa's works differently than nussl's and may produce different output.

    Returns:
        1D numpy array containing an audio signal representing the original signal used to make stft

    Example:
        
    .. code-block:: python
        :linenos:
        
        # Set up sine wave parameters
        sr = nussl.Constants.DEFAULT_SAMPLE_RATE # 44.1kHz
        n_sec = 3 # seconds
        duration = n_sec * sr
        freq = 300 # Hz

        # Make sine wave array
        x = np.linspace(0, freq * 2 * np.pi, duration)
        x = np.sin(x)

        # Set up e_stft() parameters
        win_type = nussl.WindowType.HANN
        win_length = 2048
        hop_length = win_length / 2

        # Get an stft
        stft = nussl.e_stft(x, win_length, hop_length, win_type)

        calculated_signal = nussl.e_istft(stft, win_length, hop_length)
    """

    n_hops = stft.shape[1]
    overlap = window_length - hop_length
    signal_length = (n_hops * hop_length) + overlap
    signal = np.zeros(signal_length)

    norm_window = np.zeros(signal_length)
    window = make_window(window_type, window_length)

    # Add reflection back
    stft = _add_reflection(stft) if reconstruct_reflection else stft

    for n in range(n_hops):
        start = n * hop_length
        end = start + window_length
        inv_sig_temp = np.real(scifft.ifft(stft[:, n]))
        signal[start:end] += inv_sig_temp[:window_length]
        norm_window[start:end] = norm_window[start:end] + window

    norm_window[norm_window == 0.0] = constants.EPSILON  # Prevent dividing by zero
    signal_norm = signal / norm_window

    # remove zero-padding
    if remove_padding:
        if overlap >= hop_length:
            ovp_hop_ratio = int(np.ceil(overlap / hop_length))
            start = ovp_hop_ratio * hop_length
            end = signal_length - overlap

            signal_norm = signal_norm[start:end]

        else:
            signal_norm = signal_norm[hop_length:]

    return signal_norm


def librosa_istft_wrapper(stft, window_length, hop_length, window_type,
                          remove_reflection=False, center=True, original_signal_length=None):
    """Wrapper for calling into librosa's istft function.

    Args:
        stft:
        window_length:
        hop_length:
        window_type:
        remove_reflection:
        center:
        original_signal_length:

    Returns:

    """
    window = _get_window_function(window_type) if window_type is not None else None

    if remove_reflection:
        n_fft = stft.shape[0]
        n_fft -= 1 if n_fft % 2 == 0 else 0
        stft = stft[:n_fft, :]

    signal = librosa.istft(stft, hop_length, window_length, window, center)

    # remove zero-padding
    if original_signal_length is not None:
        signal = librosa.util.fix_length(signal, original_signal_length)

    return signal


def e_stft_plus(signal, window_length, hop_length, window_type, sample_rate,
                n_fft_bins=None, remove_reflection=True):
    """
    Does a short time fourier transform (STFT) of the signal (by calling e_stft() ), but also calculates
    the power spectral density (PSD), frequency and time vectors for the calculated STFT. This function does not
    give you as many options as e_stft() (wrt removing the reflection and using librosa). If you need that
    flexibility, it is recommended that you either use e_stft() or use an AudioSignal object.

    Use this is situations where you need more than just the STFT data. For instance, this is used in plot_stft()
    to get the frequency vector to graph. In situations where you don't need this extra data it is
    more efficient to use e_stft().

    Additionally, this function assumes a single channeled audio signal and is not guaranteed to work on
    multichannel audio. If you want to do an STFT on multichannel audio see the AudioSignal object.

    Args:
        signal: 1D numpy array containing audio data. (REAL?COMPLEX?INTEGER?)
        window_length: (int) number of samples per window
        hop_length: (int) number of samples between the start of adjacent windows, or "hop"
        window_type: (string) type of window to use. Using WindowType object is recommended.
        sample_rate: (int) the intended sample rate, this is used in the calculation of the frequency vector
        n_fft_bins: (int) (Optional) number of fft bins per time window.
        If not specified, defaults to next highest power of 2 above window_length
        remove_reflection (bool):

    Returns:
        stft: (np.ndarray) a 2D matrix short time fourier transform data

    """

    if n_fft_bins is None:
        n_fft_bins = window_length

    stft = e_stft(signal, window_length, hop_length, window_type, n_fft_bins, remove_reflection)

    if remove_reflection:
        frequency_vector = (sample_rate / 2) * np.linspace(0, 1, (n_fft_bins / 2) + 1)
    else:
        frequency_vector = sample_rate * np.linspace(0, 1, n_fft_bins + 1)

    time_vector = np.array(range(stft.shape[1]))
    hop_in_secs = hop_length / (1.0 * sample_rate)
    time_vector = np.multiply(hop_in_secs, time_vector)

    window_type = constants.WINDOW_DEFAULT if window_type is None else window_type
    window = make_window(window_type, window_length)
    win_dot = np.dot(window, window.T)
    psd = np.zeros_like(stft, dtype=float)
    for i in range(psd.shape[1]):
        psd[:, i] = (1 / float(sample_rate)) * ((abs(stft[:, i]) ** 2) / float(win_dot))

    return stft, psd, frequency_vector, time_vector


def _add_zero_padding(signal, window_length, hop_length):
    """

    Args:
        signal:
        window_length:
        hop_length:
    Returns:
    """
    original_signal_length = len(signal)
    overlap = window_length - hop_length
    num_blocks = np.ceil(len(signal) / hop_length)

    if overlap >= hop_length:  # Hop is less than 50% of window length
        overlap_hop_ratio = np.ceil(overlap / hop_length)

        before = int(overlap_hop_ratio * hop_length)
        after = int((num_blocks * hop_length + overlap) - original_signal_length)

        signal = np.pad(signal, (before, after), 'constant', constant_values=(0, 0))
        extra = overlap

    else:
        after = int((num_blocks * hop_length + overlap) - original_signal_length)
        signal = np.pad(signal, (hop_length, after), 'constant', constant_values=(0, 0))
        extra = window_length

    num_blocks = int(np.ceil((len(signal) - extra) / hop_length))
    num_blocks += 1 if overlap == 0 else 0  # if no overlap, then we need to get another hop at the end

    return signal, num_blocks


def _remove_stft_padding(stft, original_signal_length, window_length, hop_length):
    """

    Args:
        stft:
        original_signal_length:
        window_length:
        hop_length:

    Returns:

    """
    overlap = window_length - hop_length
    first = int(np.ceil(overlap / hop_length))
    num_col = int(np.ceil((original_signal_length - window_length) / hop_length))
    stft_cut = stft[:, first:first + num_col]
    return stft_cut


def make_window(window_type, length, symmetric=False):
    """Returns an :obj:`np.array` populated with samples of a normalized window of type
    :param:`window_type`.

    Args:
        window_type (str): Type of window to create, string can be
        length (int): length of window
        symmetric (bool): If ``False``,  generates a periodic window (for use in spectral analysis).
            If ``True``, generates a symmetric window (for use in filter design).
            Does nothing for rectangular window.

    Returns:
        window (np.array): np array with a window of type window_type
    """

    # Generate samples of a normalized window
    if window_type == constants.WINDOW_RECTANGULAR:
        return np.ones(length)
    elif window_type == constants.WINDOW_HANN:
        return scipy.signal.hann(length, symmetric)
    elif window_type == constants.WINDOW_BLACKMAN:
        return scipy.signal.blackman(length, symmetric)
    elif window_type == constants.WINDOW_HAMMING:
        return scipy.signal.hamming(length, symmetric)
    elif window_type == constants.WINDOW_TRIANGULAR:
        return scipy.signal.triang(length, symmetric)
    else:
        return None


def _get_window_function(window_type):
    """
    Gets a window function from ``scipy.signal``
    Args:
        window_type: (string) name of the window function from scipy.signal

    Returns: callable window function from scipy.signal

    """
    if window_type in dir(scipy.signal):
        return getattr(scipy.signal, window_type)
    else:
        warnings.warn("Cannot get window type {} from scipy.signal".format(window_type))
        return None


def _add_reflection(matrix):
    reflection = matrix[-2:0:-1, :]
    reflection = reflection.conj()
    return np.vstack((matrix, reflection))


class StftParams(object):
    """
    The StftParams class is a container for information needed to run an STFT or iSTFT.
    This is meant as a convenience and does not actually perform any calculations within. It should
    get "decomposed" by the time e_stft() or e_istft() are called, so that every attribute in this
    object is a parameter to one of those functions.

    Every class that inherits from the SeparationBase class has an StftParms object, and this
    is the only way that a top level user has access to the STFT parameter settings that
    all of the separation algorithms are built upon.
    This object will get passed around instead of each of these individual attributes.
    """
    def __init__(self, sample_rate, window_length=None, hop_length=None, window_type=None, n_fft_bins=None):
        self.sample_rate = int(sample_rate)

        # default to 40ms windows
        default_win_len = int(2 ** (np.ceil(np.log2(constants.DEFAULT_WIN_LEN_PARAM * sample_rate))))
        self._window_length = default_win_len if window_length is None else int(window_length)
        self._hop_length = self._window_length // 2 if hop_length is None else int(hop_length)
        self.window_type = constants.WINDOW_DEFAULT if window_type is None else window_type
        self._n_fft_bins = self._window_length if n_fft_bins is None else int(n_fft_bins)

        self._hop_length_needs_update = True
        self._n_fft_bins_needs_update = True

        if hop_length is not None:
            self._hop_length_needs_update = False

        if n_fft_bins is not None:
            self._n_fft_bins_needs_update = False

    @property
    def window_length(self):
        return self._window_length

    @window_length.setter
    def window_length(self, value):
        """
        Length of stft window in samples. If window_overlap
        or num_fft are not set manually, then changing this will update them to
        hop_length = window_length // 2, and and num_fft = window_length
        This property is settable.
        """
        self._window_length = value

        if self._n_fft_bins_needs_update:
            self._n_fft_bins = value

        if self._hop_length_needs_update:
            self._hop_length = value // 2

    @property
    def hop_length(self):
        return self._hop_length

    @hop_length.setter
    def hop_length(self, value):
        """
        Number of samples that e_stft will jump ahead for every time slice.
        By default, this is equal to half of self.window_length and will update when you
        change self.window_length to stay at self.window_length // 2. If you set self.hop_length directly
        then self.hop_length and self.window_length are unlinked.
        This property is settable.
        """
        self._hop_length_needs_update = False
        self._hop_length = value

    @property
    def n_fft_bins(self):
        """

        Returns:

        """
        return self._n_fft_bins

    @n_fft_bins.setter
    def n_fft_bins(self, value):
        """
        Number of fft bins per time slice in the stft. A time slice is of length window length.
        By default the number of FFT bins is equal to window_length (value of window_length),
        but if this is set manually then e_stft takes a window of length.
        If you give it a value lower than self.window_length, self.window_length will be used.
        This property is settable.

        """
        # TODO: add warning for this
        if value < self.window_length:
            value = self.window_length

        self._n_fft_bins_needs_update = False
        self._n_fft_bins = value

    @property
    def window_overlap(self):
        """
        Returns number of samples of overlap between adjacent time slices.
        This is calculated like self.window_length - self.hop_length
        This property is not settable.
        """
        return self.window_length - self.hop_length

    def to_json(self):
        return json.dumps(self, default=self._to_json_helper)

    def _to_json_helper(self, o):
        if not isinstance(o, StftParams):
            raise TypeError
        d = {'__class__': o.__class__.__name__,
             '__module__': o.__module__}
        d.update(o.__dict__)
        return d

    @staticmethod
    def from_json(json_string):
        return json.loads(json_string, object_hook=StftParams._from_json_helper)

    @staticmethod
    def _from_json_helper(json_dict):
        if '__class__' in json_dict:
            class_name = json_dict.pop('__class__')
            module = json_dict.pop('__module__')
            if class_name != StftParams.__name__ or module != StftParams.__module__:
                raise TypeError
            sr = json_dict['sample_rate']
            s = StftParams(sr)
            for k, v in json_dict.items():
                s.__dict__[k] = v if not isinstance(v, unicode) else v.encode('ascii')
            return s
        else:
            return json_dict

    def __eq__(self, other):
        return all([v == other.__dict__[k] for k, v in self.__dict__.items()])

    def __ne__(self, other):
        return not self == other



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Class the will run your source separation algorithm on a number of files then evaluate using the
evaluation method of your choosing, all in one fell swoop.
"""

import warnings

import evaluation_base
from ..core import utils
from ..core import constants
from .precision_recall_fscore import PrecisionRecallFScore
from ..separation import SeparationBase, IdealMask

__all__ = ['run_and_evaluate', 'run_and_eval_prf']


def run_and_evaluate(evaluation_object, evaluation_kwargs,
                     separation_object, separation_kwargs,
                     mixture_list, true_sources_list_of_lists):
    """
    
    Args:
        evaluation_object (:class:`EvaluationBase`):
        evaluation_kwargs (dict):
        separation_object (:class:`SeparationBase`):
        separation_kwargs (dict):
        mixture_list: 
        true_sources_list_of_lists: 

    Returns:

    """

    mixture_list = utils.verify_audio_signal_list_lax(mixture_list)
    assert issubclass(separation_object, SeparationBase), 'Expected a SeparationBase derived class!'
    assert issubclass(evaluation_object, evaluation_base.EvaluationBase), \
        'Expected an EvaluationBase derived class!'

    scores = {}
    for i, mixture in enumerate(mixture_list):

        true_sources_list = utils.verify_audio_signal_list_strict(true_sources_list_of_lists[i])

        assert mixture.signal_length == true_sources_list[0].signal_length, \
            'Mixture signal_length does not match true sources at idx {}'.format(i)

        assert mixture.num_channels == true_sources_list[0].num_channels, \
            'Mixture num_channels does not match true sources at idx {}'.format(i)

        sep = separation_object(input_audio_signal=mixture, **separation_kwargs)
        sep.run()
        est_sources = sep.make_audio_signals()

        eval_ = evaluation_object(true_sources_list=true_sources_list,
                                  estimated_sources_list=est_sources)
        cur_score = eval_.evaluate()
        scores[mixture.file_name] = cur_score

    return scores


def run_and_eval_prf(separation_list, separation_kwargs,
                     mixture_list, true_sources_list_of_lists,
                     skip_errors=False, name_list=None, verbose=False):
    """
    This is a helper method to run a :ref:`MaskSeparationBase`-derived source separation algorithm
    on a list of mixtures (provided in `mixture_list`)
    Run and evaluate :ref:`PrecisionRecallFScore` for each 
    Args:
        separation_list: 
        separation_kwargs: 
        mixture_list: (list) List of :ref:`AudioSignal` objects that contain mixtures.
        true_sources_list_of_lists: (list) List of lists of AudioSignal objects that contain
            true sources
        skip_errors: (bool) 
        name_list: (list) 

    Returns:

    """
    mixture_list = utils.verify_audio_signal_list_lax(mixture_list)
    separation_list = utils.verify_mask_separation_base_list(separation_list)

    scores = {}
    for i, separation_object in enumerate(separation_list):
        for j, mixture in enumerate(mixture_list):
            if verbose:
                print('Starting {} and {}'.format(str(separation_object), mixture.file_name))

            true_sources_list = utils.verify_audio_signal_list_strict(true_sources_list_of_lists[j])

            if mixture.signal_length != true_sources_list[0].signal_length:
                error = 'Mixture signal_length does not match true sources at idx {}'.format(j)
                if skip_errors:
                    warnings.warn(error)
                    continue
                else:
                    raise RuntimeError(error)

            if mixture.num_channels != true_sources_list[0].num_channels:
                error = 'Mixture num_channels does not match true sources at idx {}'.format(j)
                if skip_errors:
                    warnings.warn(error)
                    continue
                else:
                    raise RuntimeError(error)

            # Setup and run the user provided algorithm
            sep = separation_object(input_audio_signal=mixture,
                                    mask_type=constants.BINARY_MASK,
                                    **separation_kwargs[i])

            est_mask_list = sep.run()

            # Setup and run the ideal mask
            ideal_mask_sep = IdealMask(input_audio_mixture=mixture, sources_list=true_sources_list,
                                       mask_type=constants.BINARY_MASK)

            ideal_mask_list = ideal_mask_sep.run()

            prf = PrecisionRecallFScore(true_sources_mask_list=ideal_mask_list,
                                        estimated_sources_mask_list=est_mask_list)
            cur_score = prf.evaluate()

            # TODO: This is a kludge because AudioSignal objects are not named uniquely
            score_name = mixture.file_name + '_{}'.format(id(mixture))
            scores[score_name] = cur_score

    return scores



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Provides utilities for running nussl algorithms that do not belong to
any specific algorithm or that are shared between algorithms.
"""

from __future__ import division
import warnings
import base64
import json
import re
import collections

import numpy as np
import musdb

import constants

__all__ = ['find_peak_indices', 'find_peak_values',
           'json_ready_numpy_array', 'json_serialize_numpy_array', 'load_numpy_json',
           'json_numpy_obj_hook',
           'add_mismatched_arrays', 'add_mismatched_arrays2D', 'complex_randn',
           '_get_axis',
           'print_all_separation_algorithms',
           'verify_audio_signal_list_lax', 'verify_audio_signal_list_strict',
           'verify_separation_base_list', 'verify_mask_separation_base_list',
           '_verify_audio_data', '_verify_transformation_data']


def find_peak_indices(input_array, n_peaks, min_dist=None, do_min=False, threshold=0.5):
    """
    This function will find the indices of the peaks of an input n-dimensional numpy array.
    This can be configured to find max or min peak indices, distance between the peaks, and
    a lower bound, at which the algorithm will stop searching for peaks (or upper bound if
    searching for max). Used exactly the same as :func:`find_peak_values`.

    This function currently only accepts 1-D and 2-D numpy arrays.

    Notes:
        * This function only returns the indices of peaks. If you want to find peak values,
        use :func:`find_peak_values`.

        * min_dist can be an int or a tuple of length 2.
            If input_array is 1-D, min_dist must be an integer.
            If input_array is 2-D, min_dist can be an integer, in which case the minimum
            distance in both dimensions will be equal. min_dist can also be a tuple if
            you want each dimension to have a different minimum distance between peaks.
            In that case, the 0th value in the tuple represents the first dimension, and
            the 1st value represents the second dimension in the numpy array.


    See Also:
        :: :func:`find_peak_values` ::

    Args:
        input_array: a 1- or 2- dimensional numpy array that will be inspected.
        n_peaks: (int) maximum number of peaks to find
        min_dist: (int) minimum distance between peaks. Default value: len(input_array) / 4
        do_min: (bool) if True, finds indices at minimum value instead of maximum
        threshold: (float) the value (scaled between 0.0 and 1.0)

    Returns:
        peak_indices: (list) list of the indices of the peak values

    """
    input_array = np.array(input_array, dtype=float)

    if input_array.ndim > 2:
        raise ValueError('Cannot find peak indices on data greater than 2 dimensions!')

    is_1d = input_array.ndim == 1
    zero_dist = zero_dist0 = zero_dist1 = None
    min_dist = len(input_array) // 4 if min_dist is None else min_dist

    if is_1d:
        zero_dist = min_dist
    else:
        if type(min_dist) is int:
            zero_dist0 = zero_dist1 = min_dist
        elif len(min_dist) == 1:
            zero_dist0 = zero_dist1 = min_dist[0]
        else:
            zero_dist0, zero_dist1 = min_dist

    # scale input_array between [0.0, 1.0]
    if np.min(input_array) < 0.0:
        input_array += np.min(input_array)
    elif np.min(input_array) > 0.0:
        input_array -= np.min(input_array)

    input_array /= np.max(input_array)

    # flip sign if doing min
    input_array = -input_array if do_min else input_array

    # throw out everything below threshold
    input_array = np.multiply(input_array, (input_array >= threshold))

    # check to make sure we didn't throw everything out
    if np.size(np.nonzero(input_array)) == 0:
        raise ValueError('Threshold set incorrectly. No peaks above threshold.')
    if np.size(np.nonzero(input_array)) < n_peaks:
        warnings.warn('Threshold set such that there will be less peaks than n_peaks.')

    peak_indices = []
    for i in range(n_peaks):
        # np.unravel_index for 2D indices e.g., index 5 in a 3x3 array should be (1, 2)
        # Also, wrap in list for duck typing
        cur_peak_idx = list(np.unravel_index(np.argmax(input_array), input_array.shape))

        # zero out peak and its surroundings
        if is_1d:
            cur_peak_idx = cur_peak_idx[0]
            peak_indices.append(cur_peak_idx)
            lower, upper = _set_array_zero_indices(cur_peak_idx, zero_dist, len(input_array))
            input_array[lower:upper] = 0
        else:
            peak_indices.append(cur_peak_idx)
            lower0, upper0 = _set_array_zero_indices(cur_peak_idx[0], zero_dist0,
                                                     input_array.shape[0])
            lower1, upper1 = _set_array_zero_indices(cur_peak_idx[1], zero_dist1,
                                                     input_array.shape[1])
            input_array[lower0:upper0, lower1:upper1] = 0

        if np.sum(input_array) == 0.0:
            break

    return peak_indices


def _set_array_zero_indices(index, zero_distance, max_len):
    lower = index - zero_distance - 1
    upper = index + zero_distance + 1
    lower = 0 if lower < 0 else lower
    upper = max_len if upper >= max_len else upper
    return lower, upper


def find_peak_values(input_array, n_peaks, min_dist=None, do_min=False, threshold=0.5):
    """
    Finds the values of the peaks in a 1-D or 2-D numpy array. Used exactly the same as
    :func:`find_peak_indices`. This function will find the values of the peaks of an input
    n-dimensional numpy array.

    This can be configured to find max or min peak values, distance between the peaks, and
    a lower bound, at which the algorithm will stop searching for peaks (or upper bound if
    searching for max).

    This function currently only accepts 1-D and 2-D numpy arrays.

    Notes:
        * This function only returns the indices of peaks. If you want to find peak values,
        use :func:`find_peak_indices`.

        * min_dist can be an int or a tuple of length 2.
            If input_array is 1-D, min_dist must be an integer.
            If input_array is 2-D, min_dist can be an integer, in which case the minimum
            distance in both dimensions will be equal. min_dist can also be a tuple if
            you want each dimension to have a different minimum distance between peaks.
            In that case, the 0th value in the tuple represents the first dimension, and
            the 1st value represents the second dimension in the numpy array.


    See Also:
        :: :func:`find_peak_indices` ::

    Args:
        input_array: a 1- or 2- dimensional numpy array that will be inspected.
        n_peaks: (int) maximum number of peaks to find
        min_dist: (int) minimum distance between peaks. Default value: len(input_array) / 4
        do_min: (bool) if True, finds indices at minimum value instead of maximum
        threshold: (float) the value (scaled between 0.0 and 1.0)

    Returns:
        peak_values: (list) list of the values of the peak values

    """
    if input_array.ndim > 2:
        raise ValueError('Cannot find peak indices on data greater than 2 dimensions!')

    if input_array.ndim == 1:
        return [input_array[i] for i in find_peak_indices(input_array, n_peaks, min_dist,
                                                          do_min, threshold)]
    else:
        return [input_array[i, j] for i, j in find_peak_indices(input_array, n_peaks, min_dist,
                                                                do_min, threshold)]


def json_ready_numpy_array(array):
    """
    Adapted from:
    http://stackoverflow.com/a/27948073/5768001
    Args:
        array: np array to make json ready.

    Returns:

    """
    if isinstance(array, np.ndarray):
        # noinspection PyTypeChecker
        data_b64 = base64.b64encode(np.ascontiguousarray(array).data)
        return {
                constants.NUMPY_JSON_KEY: {
                        "__ndarray__": data_b64,
                        "dtype":  str(array.dtype),
                        "shape": array.shape
                    }
                }

    return None


def json_serialize_numpy_array(array):
    """
    Returns a JSON string of the numpy array.

    Notes:
        The generated JSON strings can be converted back to numpy arrays with load_numpy_json()

    Args:
        array: (numpy array) any numpy array to convert to JSON

    Returns:
        (string) JSON-ified numpy array.

    See Also:
        load_numpy_json()
    """
    return json.dumps(json_ready_numpy_array(array))


def load_numpy_json(array_json):
    """
    Turns a JSON-ified numpy array back into a regular numpy array.

    Notes:
        This function is only guaranteed to work with JSON generated by json_serialize_numpy_array()

    Args:
        array_json: (string) JSON-ified nump array

    Returns:
        (numpy array) numpy array from the input JSON string

    See Also:
        json_serialize_numpy_array()
    """
    return json.loads(array_json, object_hook=json_numpy_obj_hook)[constants.NUMPY_JSON_KEY]


def json_numpy_obj_hook(dct):
    """
    Decodes a previously encoded numpy ndarray
    with proper shape and dtype
    from: http://stackoverflow.com/a/27948073/5768001
    :param dct: (dict) json encoded ndarray
    :return: (ndarray) if input was an encoded ndarray
    """
    if isinstance(dct, dict) and '__ndarray__' in dct:
        data = base64.b64decode(dct['__ndarray__'])
        return np.frombuffer(data, dct['dtype']).reshape(dct['shape'])
    return dct


def add_mismatched_arrays(array1, array2, truncate=False):
    """
    Will add two 1D numpy arrays of different length. If :param:`truncate` is ``False``, it will
    expand the resultant array to the larger of the two, if ``True`` it will truncate the resultant
    array to the smaller of the two.

    Args:
        array1: (:obj:`np.ndarray`) 1D numeric numpy array
        array2: (:obj:`np.ndarray`) 1D numeric numpy array
        truncate: (bool) If ``Tru``e, will truncate the resultant array to the smaller of the two

    Returns:
        (:obj:`np.ndarray`): A 1D numpy array that is the element-wise sum of the two input arrays.

    """
    # Cast these arrays to the largest common type
    array1 = np.array(array1, dtype=np.promote_types(array1.dtype, array2.dtype))
    array2 = np.array(array2, dtype=np.promote_types(array1.dtype, array2.dtype))

    # TODO: find a more elegant way to do this whole function

    if truncate:
        if len(array1) < len(array2):
            result = array1.copy()
            result += array2[:len(array1)]
        else:
            result = array2.copy()
            result += array1[:len(array2)]
    else:
        if len(array1) < len(array2):
            result = array2.copy()
            result[:len(array1)] += array1
        else:
            result = array1.copy()
            result[:len(array2)] += array2

    return result


# noinspection PyPep8Naming
def add_mismatched_arrays2D(array1, array2, truncate=False):
    """
    Will add two 2D numpy arrays of different length. If :param:`truncate` is ``False``, it will
    expand the resultant array to the larger of the two, if ``True`` it will truncate the resultant
    array to the smaller of the two.

    Args:
        array1: (:obj:`np.array`) 2D numeric numpy array.
        array2: (:obj:`np.array`) 2D numeric numpy array.
        truncate: (bool) If ``True``, will truncate the resultant array to the smaller of the two.

    Returns:
        (:obj:`np.ndarray`): A 2D numpy array that is the element-wise sum of the two input arrays.

    """
    # Cast these arrays to the largest common type
    array1 = np.array(array1, dtype=np.promote_types(array1.dtype, array2.dtype))
    array2 = np.array(array2, dtype=np.promote_types(array1.dtype, array2.dtype))

    # TODO: find a more elegant way to do this whole function

    if truncate:
        if array1.shape[1] < array2.shape[1]:  # Kludge
            result = array1.copy()
            result += array2[:, :array1.shape[1]]
        else:
            result = array2.copy()
            result += array1[:, :array2.shape[1]]
    else:
        if array1.shape[1] < array2.shape[1]:
            result = array2.copy()
            result[:, :array1.shape[1]] += array1
        else:
            result = array1.copy()
            result[:, :array2.shape[1]] += array2

    return result


def complex_randn(shape):
    """
    Returns a complex-valued numpy array of random values with shape :param:`shape`.

    Args:
        shape (tuple): Tuple of ints that will be the shape of the resultant complex numpy array.

    Returns:
        (:obj:`np.ndarray`): a complex-valued numpy array of random values with shape `shape`
    """
    return np.random.randn(*shape) + 1j * np.random.randn(*shape)


def _get_axis(array, axis_num, i):
    """
    Will get index 'i' along axis 'axis_num' of a 2- or 3-dimensional numpy array.
    If array has 4+ dimensions or 'axis_num' is larger than number of axes, will return None.

    Args:
        array (:obj:`np.ndarray`): Array to fetch axis of.
        axis_num (int): Axes to retrieve.
        i (int): Index to retrieve.

    Returns:
        The value at index :param:`i` along axis :param:`axis_num`
    """
    if array.ndim == 2:
        if axis_num == 0:
            return array[i, :]
        elif axis_num == 1:
            return array[:, i]
        else:
            return None
    elif array.ndim == 3:
        if axis_num == 0:
            return array[i, :, :]
        elif axis_num == 1:
            return array[:, i, :]
        elif axis_num == 2:
            return array[:, :, i]
        else:
            return None
    else:
        return None


def CamelCase_to_snake_case(text):
    """
    Convert a string from camel case (``StringsThatLookLikeThis``) to snake case
    (``strings_that_look_like_this``). This is useful for matching names of
    Args:
        text:

    Returns:

    """
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', text)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def _format(string):
    """ Formats a class name correctly for checking function and class names.
        Strips all non-alphanumeric chars and makes lowercase.
    """
    return str(filter(str.isalnum, string)).lower()


def audio_signals_to_musdb_track(mixture, sources_dict, targets_dict):
    """
    Converts :class:`AudioSignal` objects to ``musdb`` :class:`Track` objects that
    contain the mixture, the ground truth sources, and the targets for use with the ``mus_eval``
    implementation of BSS-Eval and ``musdb``.

    See Also:
        * More information on ``musdb``:  `Github<https://github.com/sigsep/sigsep-mus-db>`
            and `documentation<http://musdb.readthedocs.io/>`
        * More information on ``mus_eval``: `Github<https://github.com/sigsep/sigsep-mus-eval>`
            and `documentation<https://sigsep.github.io/sigsep-mus-eval/>`
        * :class:`BSSEvalV4` for ``nussl``'s interface to BSS-Eval v4.

    Examples:
        .. code-block:: python
            :linenos:
            import nussl
            signal = nussl.AudioSignal(nussl.efz_utils.download_audio_file('HistoryRepeating.wav'))

            repet = nussl.Repet(signal)
            repet.run()

            bg, fg = repet.make_audio_signals()

            src_dict = {'vocals': fg, 'accompaniment': bg}
            target = nussl.core.constants.STEM_TARGET_DICT
            track = nussl.utils.audio_signals_to_musdb_track(signal, src_dict, target)

    Args:
        mixture (:class:`AudioSignal`): The :class:`AudioSignal` object that contains the mixture.
        sources_dict (dict): Dictionary where the keys are the labels for the sources and values
            are the associated :class:`AudioSignal` objects.
        targets_dict (dict): Dictionary where the keys are the labels for the sources (as above)
            and the values are weights.

    Returns:
        (:obj:`musdb.Track`) populated as specified by inputs.
    """
    verify_audio_signal_list_strict(sources_dict.values() + [mixture])

    track = musdb.Track(mixture.file_name, is_wav=True, stem_id=0)
    track.audio = mixture.audio_data.T
    track.rate = mixture.sample_rate

    sources = {}
    i = 1
    for key, sig in list(sources_dict.items()):
        sources[key] = musdb.Source(name=key, stem_id=i, is_wav=True)
        sources[key].audio = sig.audio_data.T
        sources[key].rate = sig.sample_rate
        i += 1

    track.sources = sources

    targets = collections.OrderedDict()
    for name, target_srcs in list(targets_dict.items()):
        # add a list of target sources
        target_sources = []
        for source, gain in list(target_srcs.items()):
            if source in list(track.sources.keys()):
                # add gain to source tracks
                track.sources[source].gain = float(gain)
                # add tracks to components
                target_sources.append(sources[source])
        # add sources to target
        if target_sources:
            targets[name] = musdb.Target(sources=target_sources)

    track.targets = targets
    return track


def verify_audio_signal_list_lax(audio_signal_list):
    """
    Verifies that an input (:param:`audio_signal_list`) is a list of :ref:`AudioSignal` objects.
    If not so, attempts to correct the list (if possible) and returns the corrected list.

    Args:
        audio_signal_list (list): List of :ref:`AudioSignal` objects

    Returns:
        audio_signal_list (list): Verified list of :ref:`AudioSignal` objects.

    """
    # Lazy load to prevent a circular reference upon initialization
    from .audio_signal import AudioSignal

    if isinstance(audio_signal_list, AudioSignal):
        audio_signal_list = [audio_signal_list]
    elif isinstance(audio_signal_list, list):
        if not all(isinstance(s, AudioSignal) for s in audio_signal_list):
            raise ValueError('All input objects must be AudioSignal objects!')
        if not all(s.has_data for s in audio_signal_list):
            raise ValueError('All AudioSignal objects in input list must have data!')
    else:
        raise ValueError('All input objects must be AudioSignal objects!')

    return audio_signal_list


def verify_audio_signal_list_strict(audio_signal_list):
    """
    Verifies that an input (:param:`audio_signal_list`) is a list of :ref:`AudioSignal` objects and
    that they all have the same sample rate and same number of channels. If not true,
    attempts to correct the list (if possible) and returns the corrected list.

    Args:
        audio_signal_list (list): List of :ref:`AudioSignal` objects

    Returns:
        audio_signal_list (list): Verified list of :ref:`AudioSignal` objects, that all have
        the same sample rate and number of channels.

    """
    audio_signal_list = verify_audio_signal_list_lax(audio_signal_list)

    if not all(audio_signal_list[0].sample_rate == s.sample_rate for s in audio_signal_list):
        raise ValueError('All input AudioSignal objects must have the same sample rate!')

    if not all(audio_signal_list[0].num_channels == s.num_channels for s in audio_signal_list):
        raise ValueError('All input AudioSignal objects must have the same number of channels!')

    if not all(audio_signal_list[0].signal_length == s.signal_length for s in audio_signal_list):
        raise ValueError('All input AudioSignal objects must have the same signal length!')

    return audio_signal_list


def verify_separation_base_list(separation_list):
    """
    Verifies that all items in :param:`separation_list` are :ref:`SeparationBase` -derived objects.
    If not so, attempts to correct the list if possible and returns the corrected list.

    Args:
        separation_list (list): List of :ref:`SeparationBase` -derived objects

    Returns:
        separation_list (list): Verified list of :ref:`SeparationBase` -derived objects

    """
    # Lazy load to prevent a circular reference upon initialization
    from ..separation import SeparationBase

    if isinstance(separation_list, SeparationBase):
        separation_list = [separation_list]
    elif isinstance(separation_list, list):
        if not all(issubclass(s, SeparationBase) for s in separation_list):
            raise ValueError('All separation objects must be SeparationBase-derived objects!')
    else:
        raise ValueError('All separation objects must be SeparationBase-derived objects!')

    return separation_list


def verify_mask_separation_base_list(mask_separation_list):
    """
    Verifies that all items in :param:`separation_list` are :ref:`MaskSeparationBase` -derived
    objects. If not so, attempts to correct the list if possible and returns the corrected list.

    Args:
        mask_separation_list (list): List of :ref:`MaskSeparationBase` -derived objects

    Returns:
        separation_list (list): Verified list of :ref:`MaskSeparationBase` -derived objects

    """
    # Lazy load to prevent a circular reference upon initialization
    from ..separation import MaskSeparationBase

    if isinstance(mask_separation_list, MaskSeparationBase):
        mask_separation_list = [mask_separation_list]
    elif isinstance(mask_separation_list, list):
        if not all(issubclass(s, MaskSeparationBase) for s in mask_separation_list):
            raise ValueError('All separation objects must be '
                             'MaskSeparationBase-derived objects! {}'.format(mask_separation_list))
    else:
        raise ValueError('All separation objects must be MaskSeparationBase-derived objects!')

    return mask_separation_list


def _verify_audio_data(audio_data):
    """
    A helper method to make sure that input audio data is formatted correctly. This checks if
    :param:`audio_data` is a numpy array, then if it's all finite
    Args:
        audio_data (:obj:`np.ndarray`): A numpy array with audio-data. Can be  1 or 2 dimensional

    Returns:
        Correctly formatted :param:`audio_data` array or `None` if :param:`audio_data` is `None`

    """
    if audio_data is None:
        return None

    if isinstance(audio_data, list):
        audio_data = np.array(audio_data)

    if not isinstance(audio_data, np.ndarray):
        raise ValueError('Type of audio_data must be of type np.ndarray!')

    if not np.isfinite(audio_data).all():
        raise ValueError('Not all values of audio_data are finite!')

    if audio_data.ndim > 1 \
            and audio_data.shape[constants.CHAN_INDEX] > audio_data.shape[constants.LEN_INDEX]:
        warnings.warn('audio_data is not as we expect it. Transposing signal...')
        audio_data = audio_data.T

    if audio_data.ndim > 2:
        raise ValueError('audio_data cannot have more than 2 dimensions!')

    if audio_data.ndim < 2:
        audio_data = np.expand_dims(audio_data, axis=constants.CHAN_INDEX)

    return audio_data


def _verify_transformation_data(transformation_data):
    if transformation_data is None:
        return None

    if isinstance(transformation_data, list):
        transformation_data = np.array(transformation_data)

    if not isinstance(transformation_data, np.ndarray):
        raise ValueError('Type of transformation_data must be of type np.ndarray!')

    if transformation_data.ndim == 1:
        raise ValueError('Cannot support arrays with less than 2 dimensions!')

    if transformation_data.ndim == 2:
        transformation_data = np.expand_dims(transformation_data, axis=constants.TF_CHAN_INDEX)

    if transformation_data.ndim > 3:
        raise ValueError('Cannot support arrays with more than 3 dimensions!')

    return transformation_data


def print_all_separation_algorithms():
    """
    Displays the list of all separation algorithms in nussl to the console.
    """
    from ..separation import all_separation_algorithms
    print('\n'.join([a.__name__ for a in all_separation_algorithms]))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
from scipy.ndimage.filters import maximum_filter, minimum_filter

from ..core.audio_signal import AudioSignal
from ..core import constants
import mask_separation_base
import masks


class FT2D(mask_separation_base.MaskSeparationBase):
    """Implements foreground/background separation using the 2D Fourier Transform

    Parameters:
        input_audio_signal: (AudioSignal object) The AudioSignal object that has the
                            audio data that REPET will be run on.
        high_pass_cutoff: (Optional) (float) value (in Hz) for the high pass cutoff filter.
        do_mono: (Optional) (bool) Flattens AudioSignal to mono before running the algorithm
            (does not effect the input AudioSignal object)
        use_librosa_stft: (Optional) (bool) Calls librosa's stft function instead of nussl's

    """
    def __init__(self, input_audio_signal, high_pass_cutoff=None, neighborhood_size=(1, 25),
                 do_mono=False, use_librosa_stft=constants.USE_LIBROSA_STFT,
                 mask_type=mask_separation_base.MaskSeparationBase.SOFT_MASK):
        super(FT2D, self).__init__(input_audio_signal=input_audio_signal, mask_type=mask_type)
        self.high_pass_cutoff = 100.0 if high_pass_cutoff is None else float(high_pass_cutoff)
        self.background = None
        self.foreground = None
        self.use_librosa_stft = use_librosa_stft
        self.neighborhood_size = neighborhood_size
        self.result_masks = None

        self.stft = None
        self.ft2d = None

        if do_mono:
            self.audio_signal.to_mono(overwrite=True)

    def run(self):
        """

        Returns:
            background (AudioSignal): An AudioSignal object with repeating background in
            background.audio_data
            (to get the corresponding non-repeating foreground run self.make_audio_signals())

        Example:
             ::

        """
        # High pass filter cutoff freq. (in # of freq. bins), +1 to match MATLAB implementation
        self.high_pass_cutoff = int(np.ceil(self.high_pass_cutoff * (self.stft_params.n_fft_bins - 1) /
                                            self.audio_signal.sample_rate)) + 1

        # the MATLAB implementation had
        self._compute_spectrograms()

        # separate the mixture background by masking
        background_stft = []
        background_mask = []
        for i in range(self.audio_signal.num_channels):
            repeating_mask = self.compute_ft2d_mask(self.ft2d[:, :, i])
            repeating_mask[0:self.high_pass_cutoff, :] = 1  # high-pass filter the foreground
            background_mask.append(repeating_mask)

            # apply mask
            stft_with_mask = repeating_mask * self.stft[:, :, i]
            background_stft.append(stft_with_mask)

        background_stft = np.array(background_stft).transpose((1, 2, 0))
        self.background = AudioSignal(stft=background_stft,
                                      sample_rate=self.audio_signal.sample_rate)
        self.background.istft(self.stft_params.window_length, self.stft_params.hop_length,
                              self.stft_params.window_type,
                              overwrite=True, use_librosa=self.use_librosa_stft,
                              truncate_to_length=self.audio_signal.signal_length)

        background_mask = np.array(background_mask).transpose((1, 2, 0)).astype('float')
        background_mask = masks.SoftMask(background_mask)
        if self.mask_type == self.BINARY_MASK:
            background_mask = background_mask.mask_to_binary(self.mask_threshold)

        self.result_masks = [background_mask, background_mask.inverse_mask()]

        return self.result_masks
    
    def _compute_spectrograms(self):
        self.stft = self.audio_signal.stft(overwrite=True, remove_reflection=True,
                                           use_librosa=self.use_librosa_stft)
        self.ft2d = np.stack([np.fft.fft2(np.abs(self.stft[:, :, i]))
                              for i in range(self.audio_signal.num_channels)], axis = -1)

    def compute_ft2d_mask(self, ft2d):
        bg_ft2d, fg_ft2d = self.filter_local_maxima(ft2d)
        bg_stft = np.fft.ifft2(bg_ft2d)
        fg_stft = np.fft.ifft2(fg_ft2d)
        bg_mask = bg_stft > fg_stft
        return bg_mask

    def filter_local_maxima(self, ft2d):
        data = np.abs(np.fft.fftshift(ft2d))
        data /= np.max(data)
        threshold = np.std(data)
        
        data_max = maximum_filter(data, self.neighborhood_size)
        maxima = (data == data_max)
        data_min = minimum_filter(data, self.neighborhood_size)
        diff = ((data_max - data_min) > threshold)
        maxima[diff == 0] = 0
        maxima = np.maximum(maxima, np.fliplr(maxima), np.flipud(maxima))
        maxima = np.fft.ifftshift(maxima)
        
        background_ft2d = np.multiply(maxima, ft2d)
        foreground_ft2d = np.multiply(1 - maxima, ft2d)
        return background_ft2d, foreground_ft2d

    def make_audio_signals(self):
        """ Returns the background and foreground audio signals. You must have run FT2D.run() prior
        to calling this function. This function will return None if run() has not been called.

        Returns:
            Audio Signals (List): 2 element list.

                * bkgd: Audio signal with the calculated background track
                * fkgd: Audio signal with the calculated foreground track

        EXAMPLE:
             ::
        """
        if self.background is None:
            raise ValueError('Cannot make audio signals prior to running algorithm!')

        foreground_array = self.audio_signal.audio_data - self.background.audio_data
        self.foreground = self.audio_signal.make_copy_with_audio_data(foreground_array)
        return [self.background, self.foreground]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import copy
import inspect
import json
import warnings

import numpy as np

from ..core import utils
from ..core import audio_signal
from ..core import constants
from ..core.audio_signal import AudioSignal


class SeparationBase(object):
    """Base class for all separation algorithms in nussl.

    Do not call this. It will not do anything.

    Parameters:
        input_audio_signal (:class:`audio_signal.AudioSignal`). :class:`audio_signal.AudioSignal` object.
                            This will always make a copy of the provided AudioSignal object.
    """

    def __init__(self, input_audio_signal):
        if not isinstance(input_audio_signal, AudioSignal):
            raise ValueError('input_audio_signal is not an AudioSignal object!')

        self._audio_signal = None

        if input_audio_signal is not None:
            self.audio_signal = input_audio_signal
        else:
            self.audio_signal = AudioSignal()

        if not self.audio_signal.has_data:
            warnings.warn('input_audio_signal has no data!')

            # initialize to empty arrays so that we don't crash randomly
            self.audio_signal.audio_data = np.array([])
            self.audio_signal.stft_data = np.array([[]])

    @property
    def sample_rate(self):
        """(int): Sample rate of :attr:`audio_signal`.
        Literally :attr:`audio_signal.sample_rate`.
        """
        return self.audio_signal.sample_rate

    @property
    def stft_params(self):
        """(:class:`spectral_utils.StftParams`): :class:`spectral_utils.StftParams` of :attr:`audio_signal`
        Literally :attr:`audio_signal.stft_params`.
        """
        return self.audio_signal.stft_params

    @property
    def audio_signal(self):
        """(:class:`audio_signal.AudioSignal`): Copy of the :class:`audio_signal.AudioSignal` object passed in 
        upon initialization.
        """
        return self._audio_signal

    @audio_signal.setter
    def audio_signal(self, input_audio_signal):
        self._audio_signal = copy.copy(input_audio_signal)

    def plot(self, output_name, **kwargs):
        """Plots relevant data for separation algorithm

        Raises:
            NotImplementedError: Cannot call base class
        """
        raise NotImplementedError('Cannot call base class.')

    def run(self):
        """Runs separation algorithm

        Raises:
            NotImplementedError: Cannot call base class
        """
        raise NotImplementedError('Cannot call base class.')

    def make_audio_signals(self):
        """Makes :class:`audio_signal.AudioSignal` objects after separation algorithm is run

        Raises:
            NotImplementedError: Cannot call base class
        """
        raise NotImplementedError('Cannot call base class.')

    def to_json(self):
        """
        Outputs JSON from the data stored in this object.
        
        Returns:
            (str) a JSON string containing all of the information to restore this object exactly as it was when this
            was called.
            
        See Also:
            :func:`from_json` to restore a JSON frozen object.

        """
        return json.dumps(self, default=SeparationBase._to_json_helper)

    def __str__(self):
        return self.__class__.__name__

    @staticmethod
    def _to_json_helper(o):
        if not isinstance(o, SeparationBase):
            raise TypeError('SeparationBase._to_json_helper() got foreign object!')

        d = copy.copy(o.__dict__)
        for k, v in d.items():
            if isinstance(v, np.ndarray):
                d[k] = utils.json_ready_numpy_array(v)
            elif hasattr(v, 'to_json'):
                d[k] = v.to_json()
            elif isinstance(v, (list, tuple, set)) and any(hasattr(itm, 'to_json') for itm in v):
                s = []
                for itm in v:
                    if hasattr(itm, 'to_json'):
                        s.append(itm.to_json())
                    else:
                        s.append(itm)
                d[k] = s

        d['__class__'] = o.__class__.__name__
        d['__module__'] = o.__module__
        if 'self' in d:
            del d['self']

        return d

    @classmethod
    def from_json(cls, json_string):
        """
        Creates a new :class:`SeparationBase` object from the parameters stored in this JSON string.
        
        Args:
            json_string (str): A JSON string containing all the data to create a new :class:`SeparationBase` 
                object.

        Returns:
            (:class:`SeparationBase`) A new :class:`SeparationBase` object from the JSON string.
            
        See Also:
            :func:`to_json` to make a JSON string to freeze this object.

        """
        sep_decoder = SeparationBaseDecoder(cls)
        return sep_decoder.decode(json_string)

    def __call__(self):
        return self.run()

    def __repr__(self):
        return self.__class__.__name__ + ' instance'

    def __eq__(self, other):
        for k, v in self.__dict__.items():
            if isinstance(v, np.ndarray):
                if not np.array_equal(v, other.__dict__[k]):
                    return False
            elif k == 'self':
                pass
            elif v != other.__dict__[k]:
                return False
        return True

    def __ne__(self, other):
        return not self == other


class SeparationBaseDecoder(json.JSONDecoder):
    """ Object to decode a :class:`SeparationBase`-derived object from JSON serialization.
    You should never have to instantiate this object by hand.
    """

    def __init__(self, separation_class):
        self.separation_class = separation_class
        json.JSONDecoder.__init__(self, object_hook=self._json_separation_decoder)

    def _inspect_json_and_create_new_instance(self, json_dict):
        class_name = json_dict.pop('__class__')
        module_name = json_dict.pop('__module__')
        if class_name != self.separation_class.__name__ or module_name != self.separation_class.__module__:
            raise TypeError('Expected {}.{} but got {}.{} from json!'.format(self.separation_class.__module__,
                                                                             self.separation_class.__name__,
                                                                             module_name, class_name))

        # load the module and import the class
        module = __import__(module_name)
        class_ = getattr(module, class_name)

        if '_audio_signal' not in json_dict:
            raise TypeError('JSON string from {} does not have an AudioSignal object!'.format(class_name))

        # we know 'input_audio_signal' is always the first argument
        signal_json = json_dict.pop('_audio_signal')  # this is the AudioSignal object
        signal = AudioSignal.from_json(signal_json)

        # get the rest of the required arguments
        signature = inspect.getargspec(class_.__init__)
        # first arg is covered above (2), and we don't want the non-defaults (-len(signature.defaults))
        non_required_args = 0 if signature.defaults is None else len(signature.defaults)
        required_args = signature.args[2:-non_required_args]
        args = dict((k.encode('ascii'), json_dict[k]) for k in required_args)

        # make a new instance of separation class
        separator = class_(signal, **args)

        return json_dict, separator

    def _json_separation_decoder(self, json_dict):
        """
        Helper method for :class:`SeparationBaseDecoder`. Don't you worry your pretty little head about this.

        NEVER CALL THIS DIRECTLY!!

        Args:
            json_dict (dict): JSON dictionary provided by `object_hook`

        Returns:
            A new :class:`SeparationBase`-derived object from JSON serialization

        """
        if '__class__' in json_dict and '__module__' in json_dict:
            json_dict, separator = self._inspect_json_and_create_new_instance(json_dict)

            # fill out the rest of the fields
            for k, v in json_dict.items():
                if isinstance(v, dict) and constants.NUMPY_JSON_KEY in v:
                    separator.__dict__[k] = utils.json_numpy_obj_hook(v[constants.NUMPY_JSON_KEY])
                elif isinstance(v, (str, bytes)) and audio_signal.__name__ in v:  # TODO: test this
                    separator.__dict__[k] = AudioSignal.from_json(v)
                else:
                    separator.__dict__[k] = v if not isinstance(v, unicode) else v.encode('ascii')

            return separator
        else:
            return json_dict



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Deep Clustering modeller class
"""

from .. import torch_imported
if torch_imported:
    import torch
    import torch.nn as nn

import numpy as np


class TransformerDeepClustering(nn.Module):
    """
    Transformer Class for deep clustering
    """
    def __init__(self, hidden_size=300, input_size=150, num_layers=2, embedding_size=20):

        if not torch_imported:
            raise ImportError('Cannot import pytorch! Install pytorch to continue.')

        super(TransformerDeepClustering, self).__init__()
        self.hidden_size = hidden_size
        self.input_size = input_size
        self.embedding_size = embedding_size
        self.num_layers = num_layers
        rnn = nn.LSTM(self.input_size, self.hidden_size, self.num_layers,  bidirectional=True,
                      batch_first=True, dropout=0.5)
        linear = nn.Linear(self.hidden_size*2, self.input_size*self.embedding_size)
        self.add_module('rnn', rnn)
        self.add_module('linear', linear)

    def forward(self, input_data):
        """
        Forward training
        Args:
            input_data:

        Returns:

        """
        sequence_length = input_data.size(1)
        num_frequencies = input_data.size(2)
        output, hidden = self.rnn(input_data)
        output = output.contiguous()
        output = output.view(-1, sequence_length, 2*self.hidden_size)
        embedding = self.linear(output)
        embedding = embedding.view(-1, sequence_length*num_frequencies, self.embedding_size)
        embedding = nn.functional.normalize(embedding, p=2, dim=-1)
        return embedding

    @staticmethod
    def affinity_cost(embedding, assignments):
        """
        Function defining the affinity cost for deep clustering
        Args:
            embedding:
            assignments:

        Returns:

        """
        embedding = embedding.view(-1, embedding.size()[-1])
        assignments = assignments.view(-1, assignments.size()[-1])
        silence_mask = torch.sum(assignments, dim=-1, keepdim=True)
        embedding = silence_mask * embedding
        embedding_transpose = embedding.transpose(1, 0)
        assignments_transpose = assignments.transpose(1, 0)

        class_weights = nn.functional.normalize(torch.sum(assignments, dim=-2),
                                                p=1, dim=-1).unsqueeze(0)
        class_weights = 1.0 / (torch.sqrt(class_weights) + 1e-7)
        weights = torch.mm(assignments, class_weights.transpose(1, 0))
        assignments = assignments * weights.repeat(1, assignments.size()[-1])
        embedding = embedding * weights.repeat(1, embedding.size()[-1])

        loss_est = torch.norm(torch.mm(embedding_transpose, embedding), p=2)
        loss_est_true = 2*torch.norm(torch.mm(embedding_transpose, assignments), p=2)
        loss_true = torch.norm(torch.mm(assignments_transpose, assignments), p=2)
        loss = loss_est - loss_est_true + loss_true
        loss = loss / (loss_est + loss_true)
        return loss

    @staticmethod
    def show_model(model):
        """
        Prints a message to the console with model info
        Args:
            model:

        Returns:

        """
        print(model)
        num_parameters = 0
        for p in model.parameters():
            if p.requires_grad:
                num_parameters += np.cumprod(p.size())[-1]
        print('Number of parameters: {}'.format(num_parameters))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
EvaluationBase is the base class for all evaluation methods in nussl.
"""
from ..core import utils
from ..core.audio_signal import AudioSignal


class EvaluationBase(object):
    """
    Base class for all Evaluation classes for source separation algorithm in nussl. 
    Contains common functions for all evaluation techniques. This class should not be instantiated
    directly.
    
    Both ``true_sources_list`` and ``estimated_sources_list`` get validated using the private method
    :func:`_verify_input_list`. If your evaluation needs to verify that input is set correctly
    (recommended) overwrite that method to add checking.
    
    Args:
        true_sources_list (list): List of objects that contain one ground truth source per object.
            In some instances (such as the :class:`BSSEval` objects) this list is filled with
            :class:`AudioSignals` but in other cases it is populated with
            :class:`MaskBase` -derived objects (i.e., either a :class:`BinaryMask` or
            :class:`SoftMask` object).
        estimated_sources_list (list): List of objects that contain source estimations from a source
            separation algorithm. List should be populated with the same type of objects and in the
            same order as :param:`true_sources_list`.
        source_labels (list): List of strings that are labels for each source to be used as keys for
            the scores. Default value is `None` and in that case labels are `Source 0`, `Source 1`, etc.
        do_mono (bool): Whether to make the objects in ``true_sources_list`` and ``estimated_sources_list``
            mono prior to doing the evaluation. This may not work for all :class:`EvaluationBase`-derived objects.
        **kwargs: Additional arguments for subclasses.
    """

    def __init__(self, true_sources_list, estimated_sources_list, source_labels=None, **kwargs):
        self.true_sources_list = self._verify_input_list(true_sources_list)
        self.estimated_sources_list = self._verify_input_list(estimated_sources_list)

        if len(self.true_sources_list) != len(self.estimated_sources_list):
            raise AudioSignalListMismatchError('Must have the same number of objects in '
                                               'true_sources_list and estimated_sources_list!')

        # set the labels up correctly
        if source_labels is None:
            self.source_labels = ['Source {}'.format(i) for i in range(len(self.true_sources_list))]
        else:
            assert isinstance(source_labels, list), 'Expected source_labels to be ' \
                                                    'a list of strings!'
            if not all([isinstance(l, str) for l in source_labels]):
                raise ValueError('All labels must be strings!')

            # Can't use a source_labels list longer than the sources
            if len(source_labels) > len(self.true_sources_list):
                raise ValueError('Labels list is longer than sources list!')

            # If not all sources have labels, we'll just give them generic ones...
            if len(source_labels) < len(self.true_sources_list):
                start, stop = len(source_labels), len(self.true_sources_list)
                for i in range(start, stop):
                    source_labels.append('Source {}'.format(i))

            self.source_labels = source_labels

        self.evaluation_object_type = type(self.true_sources_list[0])
        self.do_mono = None
        self._scores = {}
        self.num_channels = self.true_sources_list[0].num_channels

        if 'do_mono' in kwargs:
            self.do_mono = kwargs['do_mono']
            if self.evaluation_object_type is AudioSignal and self.do_mono:
                self.num_channels = 1
                [t.to_mono(overwrite=True) for t in self.true_sources_list]
                [e.to_mono(overwrite=True) for e in self.estimated_sources_list]

    @staticmethod
    def _verify_input_list(audio_signal_list):
        """
        Base method for verifying a list of input objects for an :class:`EvaluationBase`-derived
        object. Override this method when creating new :class:`EvaluationBased`-derived class.
        
        By default calls :func:`nussl.utils.verify_audio_signal_list_strict`, which verifies that
        all objects in the input list are :class:`audio_signal.AudioSignal` objects with the same
        length, sample rate and have identical number of channels.
        
        Args:
            audio_signal_list (list): List of objects that contain one signal per object. In some
                instances (such as the :class:`BSSEval`) this list is filled with
                :ref:`AudioSignals` but in other cases it is populated with :class:`MaskBase`
                -derived objects (i.e., either a :class:`BinaryMask` or :class:`SoftMask`
                object). In the latter case, this method is overridden with a specific function 
                in :class:`evaluation.precision_recall_fscore.PrecisionRecallFScore`.

        Returns:
            A verified list of objects that are ready for running the evaluation method.

        """
        return utils.verify_audio_signal_list_strict(audio_signal_list)

    def evaluate(self):
        """
        This function runs the evaluation method. Do not call this directly from
        :class:`EvaluationBase`
        
        Raises:
            NotImplementedError

        """
        raise NotImplementedError('Cannot call base class!')

    @property
    def scores(self):
        """
        A dictionary that stores all scores from the evaluation method. Gets populated when
        :func:`evaluate` gets run.

        """
        return self._scores


class AudioSignalListMismatchError(Exception):
    """
    Error class for when true_sources_list and estimated_sources_list are different
    lengths.
    """
    pass